# -*- coding: utf-8 -*-
import pathlib

import pytest

from jira_assistant.excel_definition import ExcelDefinition
from jira_assistant.excel_operation import output_to_excel_file, read_excel_file
from jira_assistant.sprint_schedule import SprintScheduleStore
from tests.utils import read_stories_from_excel

from . import ASSETS_FILES, SRC_ASSETS


def test_read_excel_file():
    excel_definition = ExcelDefinition()
    excel_definition.load_file(SRC_ASSETS / "excel_definition.json")
    sprint_schedule = SprintScheduleStore()
    sprint_schedule.load_file(SRC_ASSETS / "sprint_schedule.json")

    columns, stories = read_excel_file(
        ASSETS_FILES / "excel.xlsx", excel_definition, sprint_schedule
    )
    assert len(columns) == 29
    assert len(stories) == 8


def test_read_excel_file_duplicate_column():
    excel_definition = ExcelDefinition()
    excel_definition.load_file(SRC_ASSETS / "excel_definition.json")
    sprint_schedule = SprintScheduleStore()
    sprint_schedule.load_file(SRC_ASSETS / "sprint_schedule.json")

    with pytest.raises(ValueError) as e:
        _, _ = read_excel_file(
            ASSETS_FILES / "excel_duplicate_column.xlsx",
            excel_definition,
            sprint_schedule,
        )

    assert "The input excel file has duplicate column." in str(e.value)


def test_read_excel_file_empty_row():
    excel_definition = ExcelDefinition()
    excel_definition.load_file(SRC_ASSETS / "excel_definition.json")
    sprint_schedule = SprintScheduleStore()
    sprint_schedule.load_file(SRC_ASSETS / "sprint_schedule.json")

    columns, stories = read_excel_file(
        ASSETS_FILES / "excel_with_empty_row.xlsx",
        excel_definition,
        sprint_schedule,
    )

    assert len(columns) == 29
    assert len(stories) == 2


def test_read_excel_file_missing_columns():
    excel_definition = ExcelDefinition()
    excel_definition.load_file(SRC_ASSETS / "excel_definition.json")
    sprint_schedule = SprintScheduleStore()
    sprint_schedule.load_file(SRC_ASSETS / "sprint_schedule.json")

    with pytest.raises(ValueError) as e:
        _, _ = read_excel_file(
            ASSETS_FILES / "excel_missing_columns.xlsx",
            excel_definition,
            sprint_schedule,
        )
    assert "Following columns are missing:" in str(e.value)
    assert "1.criticalDefect" in str(e.value)
    assert "3.partnerPriority" in str(e.value)


def test_output_to_excel_file(tmpdir):
    columns, stories = read_stories_from_excel(
        ASSETS_FILES / "excel.xlsx",
        SRC_ASSETS / "excel_definition.json",
        SRC_ASSETS / "sprint_schedule.json",
    )

    output_to_excel_file(tmpdir / "excel_direct_output.xlsx", stories, columns)

    assert (tmpdir / "excel_direct_output.xlsx").exists()


def test_output_to_excel_file_path_is_not_absolute():
    with pytest.raises(ValueError) as e:
        output_to_excel_file("excel_direct_output.xlsx", [], [])

    assert "The output file path is invalid." in str(e.value)


def test_output_to_excel_file_path_already_exist_over_write_is_true(tmpdir):
    columns, stories = read_stories_from_excel(
        ASSETS_FILES / "excel.xlsx",
        SRC_ASSETS / "excel_definition.json",
        SRC_ASSETS / "sprint_schedule.json",
    )

    output_file = tmpdir / "excel_direct_output.xlsx"

    pathlib.Path(output_file).resolve().touch()

    output_to_excel_file(
        output_file,
        stories,
        columns,
        over_write=True,
    )

    assert output_file.exists()


def test_output_to_excel_file_path_already_exist_over_write_is_false(tmpdir):
    output_file = tmpdir / "excel_direct_output.xlsx"

    pathlib.Path(output_file).resolve().touch()

    with pytest.raises(FileExistsError) as e:
        output_to_excel_file(
            output_file,
            [],
            [],
            over_write=False,
        )

    assert "already exist." in str(e.value)



# -*- coding: utf-8 -*-
from decimal import Decimal

from pytest import raises
from jira_assistant.excel_definition import SortStrategy

from jira_assistant.story import (
    compare_story_based_on_inline_weights,
    convert_to_bool,
    convert_to_datetime,
    convert_to_decimal,
    sort_stories_by_property_and_order,
    sort_stories_by_raise_ranking,
)
from tests.utils import (
    mock_story_data,
    mock_story_data_for_property_sorting,
    mock_story_data_for_property_sorting_consider_parent_index,
    mock_story_data_for_raise_ranking,
)


def test_compare_story_based_on_inline_weight():
    data, _ = mock_story_data()
    s_1 = data[0]
    s_2 = data[1]
    s_3 = data[2]
    s_4 = data[3]
    s_5 = data[4]
    s_6 = data[5]
    s_7 = data[6]
    s_8 = data[7]
    s_9 = data[8]
    assert compare_story_based_on_inline_weights(s_1, s_2) < 0
    assert compare_story_based_on_inline_weights(s_1, s_3) < 0
    assert compare_story_based_on_inline_weights(s_2, s_3) < 0
    assert compare_story_based_on_inline_weights(s_3, s_5) < 0
    assert compare_story_based_on_inline_weights(s_2, s_5) < 0
    assert compare_story_based_on_inline_weights(s_4, s_5) < 0
    assert compare_story_based_on_inline_weights(s_1, s_5) < 0
    assert compare_story_based_on_inline_weights(s_6, s_7) < 0
    assert compare_story_based_on_inline_weights(s_8, s_9) < 0


def test_compare_story_property():
    data, _ = mock_story_data()
    s_8 = data[7]
    s_9 = data[8]
    assert s_8["name"] < s_9["name"]
    assert s_8["productValue"] < s_9["productValue"]


def test_str():
    data, _ = mock_story_data()

    assert "s1, N/A" in str(data[0])


def test_lt_le_gt_ge_eq():
    data, _ = mock_story_data()
    s_1 = data[0]
    s_2 = data[1]

    with raises(TypeError):
        assert s_1 < s_2
    with raises(TypeError):
        assert s_1 <= s_2
    with raises(TypeError):
        assert s_1 > s_2
    with raises(TypeError):
        assert s_1 >= s_2
    with raises(TypeError):
        assert s_1 == s_2


def test_convert_to_bool_using_correct_type():
    assert convert_to_bool(True) is True


def test_convert_to_decimal():
    assert Decimal.compare(convert_to_decimal(Decimal(1.2)), Decimal(1.2)) == Decimal(
        "0"
    )
    assert Decimal.compare(convert_to_decimal("1.2"), Decimal("1.2")) == Decimal("0")
    assert Decimal.compare(convert_to_decimal("good"), Decimal(0)) == Decimal("0")


def test_convert_to_datetime():
    assert convert_to_datetime(None) is None


def test_sort_stories_by_property_and_order():
    stories, factory = mock_story_data_for_property_sorting()

    sort_stories_by_property_and_order(stories, factory.columns, None)

    assert len(stories) == 2
    assert stories[0]["name"] == "s2"
    assert stories[1]["name"] == "s1"


def test_sort_stories_by_property_and_order_consider_parent_index():
    stories, factory = mock_story_data_for_property_sorting_consider_parent_index()

    sort_stories_by_property_and_order(
        stories,
        factory.columns,
        SortStrategy("SortOrder", True, 1, {"ParentScopeIndexRange": "2, 5"}),
    )

    assert len(stories) == 2
    assert stories[0]["name"] == "s1"
    assert stories[1]["name"] == "s2"


def test_sort_stories_by_raise_ranking():
    stories, factory = mock_story_data_for_raise_ranking()

    result = sort_stories_by_raise_ranking(stories, factory.columns, None)

    assert len(result) == 2
    assert result[0]["name"] == "s2"
    assert result[1]["name"] == "s1"



# -*- coding: utf-8 -*-
from subprocess import CalledProcessError, run

import pytest

from . import ASSETS_ENV_FILES, ASSETS_FILES


def test_process_excel_file(tmpdir):
    result = run(
        [
            "process-excel-file",
            ASSETS_FILES / "excel.xlsx",
            "--output-folder",
            tmpdir,
            "--excel-definition-file",
            ASSETS_FILES / "excel_definition_avoid_jira_operations.json",
            "--sprint-schedule-file",
            ASSETS_FILES / "sprint_schedule.json",
        ],
        capture_output=True,
        check=True,
    )

    assert "xlsx has been saved" in result.stdout.decode("utf-8")
    assert (tmpdir / "excel_sorted.xlsx").exists()


def test_process_excel_file_output_version():
    result = run(
        ["process-excel-file", "--version"],
        capture_output=True,
        check=True,
    )

    assert "process-excel-file" in result.stdout.decode("utf-8")


def test_process_excel_file_apply_env_file(tmpdir):
    result = run(
        [
            "process-excel-file",
            ASSETS_FILES / "excel.xlsx",
            "--output_folder",
            tmpdir,
            "--excel_definition_file",
            ASSETS_FILES / "excel_definition_avoid_jira_operations.json",
            "--sprint_schedule_file",
            ASSETS_FILES / "sprint_schedule.json",
            "--env_file",
            ASSETS_ENV_FILES / "default.env",
        ],
        capture_output=True,
        check=True,
    )

    assert "xlsx has been saved" in result.stdout.decode("utf-8")
    assert (tmpdir / "excel_sorted.xlsx").exists()


def test_process_excel_file_apply_jira_cloud_env_file(tmpdir):
    result = run(
        [
            "process-excel-file",
            ASSETS_FILES / "excel.xlsx",
            "--output_folder",
            tmpdir,
            "--excel_definition_file",
            ASSETS_FILES / "excel_definition_avoid_jira_operations.json",
            "--sprint_schedule_file",
            ASSETS_FILES / "sprint_schedule.json",
            "--env_file",
            ASSETS_ENV_FILES / "default_cloud.env",
        ],
        capture_output=True,
        check=True,
    )

    assert "xlsx has been saved" in result.stdout.decode("utf-8")
    assert (tmpdir / "excel_sorted.xlsx").exists()


def test_process_excel_file_run_twice(tmpdir):
    result = run(
        [
            "process-excel-file",
            ASSETS_FILES / "excel.xlsx",
            "--excel_definition_file",
            ASSETS_FILES / "excel_definition_avoid_jira_operations.json",
            "--sprint_schedule_file",
            ASSETS_FILES / "sprint_schedule.json",
            "-o",
            tmpdir,
        ],
        capture_output=True,
        check=True,
    )

    assert "xlsx has been saved" in result.stdout.decode("utf-8")
    assert (tmpdir / "excel_sorted.xlsx").exists()

    result = run(
        [
            "process-excel-file",
            ASSETS_FILES / "excel.xlsx",
            "--excel_definition_file",
            ASSETS_FILES / "excel_definition_avoid_jira_operations.json",
            "--sprint_schedule_file",
            ASSETS_FILES / "sprint_schedule.json",
            "--output_folder",
            tmpdir,
        ],
        capture_output=True,
        check=True,
    )

    assert "xlsx has been saved" in result.stdout.decode("utf-8")
    assert (tmpdir / "excel_sorted_1.xlsx").exists()


def test_process_excel_file_using_invalid_definition_file():
    with pytest.raises(CalledProcessError) as e:
        run(
            [
                "process-excel-file",
                ASSETS_FILES / "excel.xlsx",
                "--excel_definition_file",
                ASSETS_FILES / "excel_definition_invalid_json.txt",
                "--sprint_schedule_file",
                ASSETS_FILES / "sprint_schedule.json",
            ],
            capture_output=True,
            check=True,
        )

    assert "The structure of excel definition file is wrong" in str(e.value.output)


def test_process_excel_file_using_wrong_input_file():
    with pytest.raises(CalledProcessError) as e:
        run(
            [
                "process-excel-file",
                ASSETS_FILES / "excel_definition_invalid_index.json",
                "--excel_definition_file",
                ASSETS_FILES / "excel_definition_invalid_json.txt",
            ],
            capture_output=True,
            check=True,
        )

    assert "Please provide an Excel file" in str(e.value.output)


def test_process_excel_file_input_file_not_found():
    with pytest.raises(CalledProcessError) as e:
        run(
            [
                "process-excel-file",
                ASSETS_FILES / "not_found.xlsx",
                "--excel_definition_file",
                ASSETS_FILES / "excel_definition_invalid_json.txt",
            ],
            capture_output=True,
            check=True,
        )

    assert "Input file is not exist" in str(e.value.output)


def test_process_excel_file_output_folder_not_exist(tmpdir):
    tmpdir.remove(ignore_errors=True)

    result = run(
        [
            "process-excel-file",
            ASSETS_FILES / "excel.xlsx",
            "--output_folder",
            tmpdir,
            "--excel_definition_file",
            ASSETS_FILES / "excel_definition_avoid_jira_operations.json",
            "--sprint_schedule_file",
            ASSETS_FILES / "sprint_schedule.json",
        ],
        capture_output=True,
        check=True,
    )

    assert "xlsx has been saved" in result.stdout.decode("utf-8")


def test_generate_template_excel_definition(tmpdir):
    result = run(
        ["generate-template", "--output_folder", tmpdir, "excel-definition"],
        capture_output=True,
        check=True,
    )

    assert "Generate success" in result.stdout.decode("utf-8")
    assert "excel-definition" in result.stdout.decode("utf-8")


def test_generate_template_output_version():
    result = run(
        ["generate-template", "--version"],
        capture_output=True,
        check=True,
    )

    assert "generate-template" in result.stdout.decode("utf-8")


def test_generate_template_excel(tmpdir):
    result = run(
        ["generate-template", "--output_folder", tmpdir, "excel"],
        capture_output=True,
        check=True,
    )

    assert "Generate success" in result.stdout.decode("utf-8")
    assert "excel" in result.stdout.decode("utf-8")


def test_generate_template_sprint_schedule(tmpdir):
    result = run(
        ["generate-template", "sprint-schedule", "--output_folder", tmpdir],
        capture_output=True,
        check=True,
    )

    assert "Generate success" in result.stdout.decode("utf-8")
    assert "schedule" in result.stdout.decode("utf-8")


def test_generate_template_jira_field_mapping_failed(tmpdir):
    with pytest.raises(CalledProcessError) as e:
        run(
            [
                "generate-template",
                "--output_folder",
                tmpdir,
                "jira-field-mapping",
                "--env_file",
                ASSETS_ENV_FILES / "default.env",
            ],
            capture_output=True,
            check=True,
        )

    assert "Generate failed! Template type: jira-field-mapping." in str(e.value.stdout)


def test_generate_template_failed():
    with pytest.raises(CalledProcessError) as e:
        run(["generate-template", "abc"], capture_output=True, check=True)

    assert "argument template_type: invalid choice" in str(e.value.stderr)


def test_update_jira_info():
    result = run(
        [
            "update-jira-info",
            "--access_token",
            "123",
            "--url",
            "http://localhost",
            "--user_email",
            "sharry.xu@outlook.com",
            "--env_file",
            ASSETS_ENV_FILES / "default.env",
        ],
        capture_output=True,
        check=True,
    )

    assert "Add/Update jira url success" in result.stdout.decode("utf-8")
    assert "Add/Update jira access token success" in result.stdout.decode("utf-8")
    assert "Add/Update jira user email success" in result.stdout.decode("utf-8")


def test_update_jira_info_output_version():
    result = run(
        ["update-jira-info", "--v"],
        capture_output=True,
        check=True,
    )

    assert "update-jira-info" in result.stdout.decode("utf-8")


def test_update_jira_info_failed():
    with pytest.raises(CalledProcessError) as e:
        run(
            ["update-jira-info", "--access_token", " "],
            capture_output=True,
            check=True,
        )

    assert "Please check the access token" in str(e.value.output)


def test_dry_run_steps_and_sort_excel_file(tmpdir):
    result = run(
        [
            "process-excel-file",
            ASSETS_FILES / "excel.xlsx",
            "--output-folder",
            tmpdir,
            "--excel-definition-file",
            ASSETS_FILES / "excel_definition_avoid_jira_operations.json",
            "--sprint-schedule-file",
            ASSETS_FILES / "sprint_schedule.json",
            "--dry-run",
        ],
        capture_output=True,
        check=True,
    )

    output = result.stdout.decode("utf-8")
    assert not (tmpdir / "excel_sorted.xlsx").exists()
    assert "Using custom sprint schedule..." in output
    assert "Using custom excel definition..." in output
    assert "Validating excel definition success." in output
    assert "There are 29 columns in the excel." in output
    assert "There are 25 columns in the definition file." in output
    assert "There are 8 stories can be sorted." in output
    assert "Pre-process steps:" in output
    assert "FilterOutStoryWithoutId" in output
    assert "Sort strategies:" in output
    assert "InlineWeights" in output
    assert "SortOrder" in output
    assert "RaiseRanking" in output



# -*- coding: utf-8 -*-
from __future__ import annotations

import re

from pathlib import Path
from typing import Dict, Union, Set, Optional

_INDEX_RANGE_RULE = re.compile(
    r"(^( *\d+ *,)* *\d+ *$)|^ *\d+ *- *\d+ *$"
)


def is_absolute_path_valid(path: Union[str, Path]) -> bool:
    if path is None or not Path(path).is_absolute() or not Path(path).exists():
        return False
    return True


def standardize_column_name(name: str) -> "str":
    return strip_lower(name.replace("\n", "").replace(" ", ""))


def strip_lower(statement: str) -> "str":
    return statement.strip().lower()


def is_index_range_valid(statement: Optional[str]) -> bool:
    if (
        statement is None
        or not statement.strip()
        or _INDEX_RANGE_RULE.fullmatch(statement) is None
    ):
        return False
    return True


def parse_index_range(statement: Optional[str]) -> "Optional[Set[int]]":
    if statement is None or not is_index_range_valid(statement):
        return None
    if "-" in statement:
        begin = int(statement.split("-")[0])
        end = int(statement.split("-")[1])
        if begin < end:
            return set(i for i in range(begin, end + 1))
        return set(i for i in range(end, begin + 1))
    return set(int(i) for i in statement.split(","))


def dict_has_key(dictionary: Dict, key: str) -> bool:
    if strip_lower(key) in [strip_lower(i) for i in dictionary.keys()]:
        return True
    return False



# -*- coding: utf-8 -*-
from pytest import raises
from jira_assistant.sprint_schedule import SprintScheduleStore

from . import ASSETS_FILES


def test_load_sprint_schedule():
    schedule_filename = ASSETS_FILES / "sprint_schedule.json"
    store = SprintScheduleStore()
    store.load_file(schedule_filename)
    assert store.total_count() > 0


def test_get_priority():
    schedule_filename = ASSETS_FILES / "sprint_schedule.json"
    store = SprintScheduleStore()
    assert store.get_priority("R138") == 0

    store.load_file(schedule_filename)
    assert store.get_priority("R140") == 2


def test_load_file_failed():
    with raises(FileNotFoundError) as err:
        SprintScheduleStore().load_file(ASSETS_FILES / "not_exist_file.json")

    assert (
        """Please make sure the sprint schedule file \
exist and the path should be absolute."""
        in err.value.args[0]
    )


def test_load_sprint_schedule_invalid_file():
    schedule_filename = ASSETS_FILES / "sprint_schedule_invalid_key.json"
    store = SprintScheduleStore()
    store.load_file(schedule_filename)
    assert store.total_count() == 2
    assert store.get_priority("R142") == 4
    assert store.get_priority("M126") == 4


def test_load_sprint_schedule_invalid_json():
    with raises(SyntaxError) as err:
        with open(
            ASSETS_FILES / "sprint_schedule_invalid_json.txt",
            mode="r",
            encoding="utf-8",
        ) as file:
            SprintScheduleStore().load(file.read())

    assert (
        """The structure of sprint schedule file is wrong. \
Hint: Expecting property name enclosed in double quotes in line 4:9."""
        in err.value.args[0]
    )



# -*- coding: utf-8 -*-
"""This module will provide console command signatures."""
import pathlib
import warnings
from json import dump
from os import environ, remove
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

from dotenv import load_dotenv
from termcolor import cprint
from urllib3 import disable_warnings

from .excel_definition import ExcelDefinition, ExcelDefinitionColumn
from .excel_operation import output_to_excel_file, read_excel_file
from .jira_client import JiraClient, JiraField, JiraIssueType, is_jira_cloud_url
from .sprint_schedule import SprintScheduleStore
from .story import (
    Story,
    sort_stories_by_inline_weights,
    sort_stories_by_property_and_order,
    sort_stories_by_raise_ranking,
)
from .utils import standardize_column_name, strip_lower

__all__ = [
    "run_steps_and_sort_excel_file",
    "generate_jira_field_mapping_file",
    "dry_run_steps_and_sort_excel_file",
]

# Currently, the openpyxl package will report an obsolete warning.
warnings.simplefilter(action="ignore", category=UserWarning)
# Disable the HTTPS certificate verification warning.
disable_warnings()

HERE = pathlib.Path(__file__).resolve().parent
ASSETS = HERE / "assets"
DEFAULT_JSON_IDENT = 4
DEFAULT_SPRINT_SCHEDULE_FILE = ASSETS / "sprint_schedule.json"
DEFAULT_EXCEL_DEFINITION_FILE = ASSETS / "excel_definition.json"


def __clear_env_variables():
    if "JIRA_URL" in environ:
        del environ["JIRA_URL"]
    if "JIRA_ACCESS_TOKEN" in environ:
        del environ["JIRA_ACCESS_TOKEN"]
    if "JIRA_USER_EMAIL" in environ:
        del environ["JIRA_USER_EMAIL"]


def __get_jira_client(env_file: Optional[Path] = None) -> Optional[JiraClient]:
    if env_file is None:
        if not load_dotenv(ASSETS / ".env"):
            cprint(
                """The default env file is missing.
Please use the update-jira-info command to 
run any command for creating the file.""",
                color="light_yellow",
            )
            return None
    else:
        __clear_env_variables()
        if not load_dotenv(env_file):
            cprint(
                "The env file is invalid. Please double check the env file.",
                color="light_red",
            )
            return None

    jira_url: Optional[str] = environ.get("JIRA_URL", default=None)
    if jira_url is None or jira_url.isspace() or len(jira_url) == 0:
        cprint(
            """The jira url is invalid.
Please use the update-jira-info command to add/update url.""",
            color="light_red",
        )
        return None

    jira_access_token = environ.get("JIRA_ACCESS_TOKEN", default=None)
    if (
        jira_access_token is None
        or jira_access_token.isspace()
        or len(jira_access_token) == 0
    ):
        cprint(
            """The jira access token is invalid.
Please use the update-jira-info command to add/update token.""",
            color="light_red",
        )
        return None

    jira_user_email = environ.get("JIRA_USER_EMAIL", default=None)
    if is_jira_cloud_url(jira_url) and (
        jira_user_email is None
        or jira_user_email.isspace()
        or len(jira_user_email) == 0
    ):
        cprint(
            """The jira user email is invalid.
Please use the update-jira-info command to add/update user email.""",
            color="light_red",
        )
        return None

    jira_timeout: Optional[float] = None
    tmp = environ.get("JIRA_TIMEOUT", default=None)
    if tmp is not None:
        jira_timeout = float(tmp)

    jira_client = JiraClient(
        jira_url, jira_access_token, user_email=jira_user_email, timeout=jira_timeout
    )

    if not jira_client.health_check():
        cprint(
            """The jira access token is revoked.
Please use the update-jira-info command to add/update token.""",
            color="light_red",
        )
        return None

    cprint(f"Jira link: {jira_url}", color="light_cyan")

    return jira_client


def __query_jira_information(
    stories: List[Story],
    excel_definition: ExcelDefinition,
    env_file: Optional[Path] = None,
) -> bool:
    jira_client = __get_jira_client(env_file)

    if jira_client is None:
        return False

    jira_fields = []

    for definition_column in excel_definition.get_columns():
        if (
            not definition_column["query_jira_info"]
            or definition_column["jira_field_mapping"] is None
        ):
            continue
        jira_fields.append(
            {
                "name": definition_column["name"],
                # Jira API not support specific property level.
                "jira_name": definition_column["jira_field_mapping"]["name"],
                "jira_path": definition_column["jira_field_mapping"]["path"],
            }
        )

    jira_query_result = jira_client.get_stories_detail(
        [story["storyid"].strip() for story in stories], jira_fields
    )

    for story in stories:
        story_id: str = strip_lower(story["storyid"])
        if story_id in jira_query_result:
            for jira_field in jira_fields:
                story[jira_field["name"]] = jira_query_result[story_id].get(
                    jira_field["jira_path"], None
                )
        else:
            # Story ID has been changed because of convertion.
            tmp_res = jira_client.get_stories_detail([story_id], jira_fields)
            if len(tmp_res) > 0:
                story["storyid"] = list(tmp_res.keys())[0].upper()
                for jira_field in jira_fields:
                    story[jira_field["name"]] = (
                        list(tmp_res.values())[0]
                        .get(jira_field["jira_path"], "")
                        .upper()
                    )
                cprint(
                    f"Story id has been changed. \
Previous: {story_id.upper()}, \
Current: {story['storyid'].upper()}",
                    color="light_blue",
                )
            else:
                cprint(
                    f"Cannot find related information for story: {story_id}",
                    color="light_yellow",
                )
                story.need_sort = False
                continue

    return True


def __check_allowed_value(
    current_value: Any,
    jira_field: JiraField,
    excel_column: ExcelDefinitionColumn,
    project_type_name: str,
    issue_type_name: str,
) -> "bool":
    if excel_column["jira_field_mapping"] is not None:
        jira_field_path = excel_column["jira_field_mapping"]["path"]
        if not jira_field.is_value_allowed(str(current_value), jira_field_path):
            cprint(
                f"{excel_column['name']} has not allowed value: {current_value}. ProjectType: {project_type_name} and IssueType: {issue_type_name}.",  # pylint: disable=line-too-long
                color="light_yellow",
            )
            cprint("Allowed values:", color="light_yellow")
            for index, value in enumerate(jira_field.allowed_values[jira_field_path]):
                cprint(f"{index + 1}. {value}", color="light_yellow")
            return False
    return True


def __assign_new_story_field(
    is_array: bool,
    new_story_fields: Dict[str, Any],
    excel_column: ExcelDefinitionColumn,
    value: Any,
):
    if excel_column is None or excel_column["jira_field_mapping"] is None:
        return
    jira_field_path = excel_column["jira_field_mapping"]["path"]
    if is_array:
        new_story_fields[jira_field_path] = str(value).split(excel_column["delimiter"])
    else:
        new_story_fields[jira_field_path] = value


def __create_jira_stories(
    stories: List[Story],
    excel_definition: ExcelDefinition,
    env_file: Optional[Path] = None,
) -> bool:
    jira_client = __get_jira_client(env_file)

    if jira_client is None:
        return False

    for story in stories:
        input_project_type: Optional[str] = story["projecttype"]
        input_issue_type: Optional[str] = story["issuetype"]
        # Validation
        if input_project_type is None or input_issue_type is None:
            cprint(
                f"Please fulfill ProjectType/IssueType field. \
Excel row number: {story.excel_row_index}.",
                color="light_yellow",
            )
            continue
        project_type = jira_client.get_project_by_project_name(input_project_type)
        if project_type is None:
            cprint(
                f"ProjectType: {input_project_type} is not supported.",
                color="light_yellow",
            )
            continue
        issue_type: Optional[JiraIssueType] = None
        for item in jira_client.get_issue_types(project_type.name):
            if item.name.lower() == input_issue_type.lower():
                issue_type = item
                break
        if issue_type is None:
            cprint(
                f"IssueType: {input_issue_type} is not supported.", color="light_yellow"
            )
            continue

        (
            required_fields,
            not_required_fields,
        ) = jira_client.get_fields_by_project_id_and_issue_id(
            project_type.id_, issue_type.id_
        )

        new_story_fields: Dict[str, Any] = {}

        # Required fields
        all_fields_valid: bool = True
        for required_field in required_fields:
            excel_columns = excel_definition.get_column_by_jira_field_mapping_name(
                required_field.id_
            )

            if not excel_columns:
                cprint(
                    f"Excel definition missing required field: {required_field.id_}.",  # pylint: disable=line-too-long
                    color="light_red",
                )
                all_fields_valid = False
                continue

            for excel_column in excel_columns:
                excel_column_name = excel_column["name"]
                if not hasattr(story, standardize_column_name(excel_column_name)):
                    cprint(
                        f"Story missing required field: {excel_column_name}.",
                        color="light_red",
                    )
                    all_fields_valid = False
                    continue
                current_value = story[standardize_column_name(excel_column_name)]
                if not __check_allowed_value(
                    current_value,
                    required_field,
                    excel_column,
                    project_type.name,
                    input_issue_type,
                ):
                    all_fields_valid = False
                    continue
                __assign_new_story_field(
                    required_field.is_array,
                    new_story_fields,
                    excel_column,
                    current_value,
                )

        # Not required fields
        for not_required_field in not_required_fields:
            excel_columns = excel_definition.get_column_by_jira_field_mapping_name(
                not_required_field.id_
            )
            for excel_column in excel_columns:
                excel_column_name = excel_column["name"]
                if not hasattr(story, standardize_column_name(excel_column_name)):
                    continue
                is_array = not_required_field.is_array
                current_value = story[standardize_column_name(excel_column_name)]
                if current_value is None:
                    continue
                if not __check_allowed_value(
                    current_value,
                    not_required_field,
                    excel_column,
                    project_type.name,
                    input_issue_type,
                ):
                    all_fields_valid = False
                    continue
                __assign_new_story_field(
                    is_array, new_story_fields, excel_column, current_value
                )

        if not all_fields_valid:
            cprint("Please fix the above issues.", color="light_red")
            continue

        # Special fields
        new_story_fields["project.id"] = project_type.id_

        new_story = jira_client.create_story(new_story_fields)
        if new_story is not None:
            story["storyid"] = new_story.key
            cprint(
                f"New story: {jira_client.get_jira_browser_link(new_story.key)}",
                color="light_green",
            )
            continue
        return False
    return True


def __run_pre_steps(
    stories: List[Story],
    excel_definition: ExcelDefinition,
    env_file: Optional[Path] = None,
):
    # Execute pre-process steps
    pre_process_steps = excel_definition.get_pre_process_steps()

    for pre_process_step in pre_process_steps:
        cprint(f"Executing step: {pre_process_step.name}...")
        if strip_lower(pre_process_step.name) == strip_lower("RetrieveJiraInformation"):
            need_call_jira_api: bool = False
            for excel_definition_column in excel_definition.get_columns():
                if excel_definition_column["jira_field_mapping"] is not None:
                    need_call_jira_api = True
                    break

            if need_call_jira_api:
                stories_need_call_jira: List[Story] = []
                for story in stories:
                    if story.need_sort and story["storyid"] is not None:
                        stories_need_call_jira.append(story)
                if not __query_jira_information(
                    stories_need_call_jira, excel_definition, env_file
                ):
                    cprint("Retrieve jira information failed.", color="light_yellow")
                    return
        elif strip_lower(pre_process_step.name) == strip_lower(
            "FilterOutStoryWithoutId"
        ):
            for story in stories:
                if story["storyid"] is None:
                    story.need_sort = False
        elif strip_lower(pre_process_step.name) == strip_lower(
            "FilterOutStoryBasedOnJiraStatus"
        ):
            for story in stories:
                if (
                    story["status"] is not None
                    and pre_process_step.config is not None
                    and story["status"].upper()
                    in pre_process_step.config.get("JiraStatuses", [])
                ):
                    story.need_sort = False
        elif strip_lower(pre_process_step.name) == strip_lower("CreateJiraStory"):
            stories_need_to_create: List[Story] = []
            for story in stories:
                if story["storyid"] is None or story["storyid"].strip() == "":
                    stories_need_to_create.append(story)

            if stories_need_to_create:
                if not __create_jira_stories(
                    stories_need_to_create, excel_definition, env_file
                ):
                    cprint(
                        "Error occurred when creating Jira stories.",
                        color="light_yellow",
                    )
                    return
        cprint("Executing finish.")


def __run_sort_logics(
    stories: List[Story], excel_definition: ExcelDefinition
) -> Tuple[List[Story], List[Story]]:
    stories_no_need_sort = []
    stories_need_sort = []

    for story in stories:
        if story.need_sort:
            stories_need_sort.append(story)
        else:
            stories_no_need_sort.append(story)

    # Execute sorting logic.
    sort_strategies = excel_definition.get_sort_strategies()

    for sort_strategy in sort_strategies:
        cprint(f"Executing {sort_strategy.name} sorting...")
        if strip_lower(sort_strategy.name) == strip_lower("InlineWeights"):
            stories_need_sort = sort_stories_by_inline_weights(stories_need_sort)
        elif strip_lower(sort_strategy.name) == strip_lower("SortOrder"):
            sort_stories_by_property_and_order(
                stories_need_sort,
                excel_definition.get_columns(),
                sort_strategy,
            )
        elif strip_lower(sort_strategy.name) == strip_lower("RaiseRanking"):
            stories_need_sort = sort_stories_by_raise_ranking(
                stories_need_sort,
                excel_definition.get_columns(),
                sort_strategy,
            )
        cprint("Executing finish.")

    return stories_need_sort, stories_no_need_sort


def __pre_parse_excel_file(
    input_file: Union[str, Path],
    excel_definition_file: Optional[Union[str, Path]] = None,
    sprint_schedule_file: Optional[Union[str, Path]] = None,
) -> Tuple[ExcelDefinition, Optional[List[str]], Optional[List[Story]]]:
    sprint_schedule = SprintScheduleStore()
    if sprint_schedule_file is None:
        cprint("Using default sprint schedule...")
        sprint_schedule.load(DEFAULT_SPRINT_SCHEDULE_FILE.read_text(encoding="utf-8"))
    else:
        cprint("Using custom sprint schedule...")
        sprint_schedule.load_file(sprint_schedule_file)

    excel_definition = ExcelDefinition()
    if excel_definition_file is None:
        cprint("Using default excel definition...")
        excel_definition.load(DEFAULT_EXCEL_DEFINITION_FILE.read_text(encoding="utf-8"))
    else:
        cprint("Using custom excel definition...")
        excel_definition.load_file(excel_definition_file)

    validation_result = excel_definition.validate()
    if len(validation_result) != 0:
        cprint(
            """Validating excel definition failed.
Please check below information to fix first.""",
            color="light_red",
        )
        for index, item in enumerate(validation_result):
            cprint(f"{index + 1}. {item}", color="light_yellow")
        return excel_definition, None, None
    cprint("Validating excel definition success.", color="light_green")

    excel_columns, stories = read_excel_file(
        input_file, excel_definition, sprint_schedule
    )

    return excel_definition, excel_columns, stories


def run_steps_and_sort_excel_file(
    input_file: Union[str, Path],
    output_file: Union[str, Path],
    excel_definition_file: Optional[Union[str, Path]] = None,
    sprint_schedule_file: Optional[Union[str, Path]] = None,
    over_write: bool = True,
    env_file: Optional[Path] = None,
):
    """
    Sort the Excel file and output the result

    parm input_file:
        The Excel file need to be sorted. (Absolute path only)

    parm output_file:
        The sorted Excel file location. (Absolute path only)

    parm sprint_schedule_file:
        The JSON file which contains the priority list to
        calculate the :py:class:`Milestone`

    parm excel_definition_file:
        The JSON file which contains the input Excel file's structure.

    parm over_write:
        Whether the exist output file will be over-write.
    """
    excel_definition, excel_columns, stories = __pre_parse_excel_file(
        input_file, excel_definition_file, sprint_schedule_file
    )

    if stories is None or excel_columns is None:
        cprint("Parse excel file failed.", color="light_red")
        return

    if not stories:
        cprint("There are no stories inside the excel file.", color="light_yellow")
        return

    __run_pre_steps(stories, excel_definition, env_file)

    stories_need_sort, stories_no_need_sort = __run_sort_logics(
        stories, excel_definition
    )

    output_to_excel_file(
        output_file,
        stories_need_sort + stories_no_need_sort,  # First output the sorted stories.
        excel_column_names=excel_columns,
        over_write=over_write,
    )

    cprint(f"{output_file} has been saved.", color="light_green")


def generate_jira_field_mapping_file(
    file: Union[str, Path], over_write: bool = True, env_file: Optional[Path] = None
) -> bool:
    """Generate jira field mapping"""
    jira_client = __get_jira_client(env_file)

    if jira_client is None:
        return False

    output_file_path: Path = Path(file).absolute()

    if output_file_path.exists():
        if over_write:
            try:
                remove(file)
            except PermissionError as e:
                raise FileExistsError(
                    f"The exist jira field mapping file: {file} cannot be removed."
                ) from e
        else:
            raise FileExistsError(
                f"The jira field mapping file: {file} is already exist."
            )

    with open(output_file_path, mode="x", encoding="utf-8") as output_file:
        dump(
            jira_client.get_all_fields(),
            output_file,
            default=lambda o: o.__dict__,
            indent=DEFAULT_JSON_IDENT,
            sort_keys=True,
        )
        output_file.flush()
        output_file.close()

    return True


def dry_run_steps_and_sort_excel_file(
    input_file: Union[str, Path],
    excel_definition_file: Optional[Union[str, Path]] = None,
    sprint_schedule_file: Optional[Union[str, Path]] = None,
):
    """
    Analyze the input and print the result.

    parm input_file:
        The Excel file need to be sorted. (Absolute path only)

    parm output_file:
        The sorted Excel file location. (Absolute path only)

    parm sprint_schedule_file:
        The JSON file which contains the priority list to
        calculate the :py:class:`Milestone`

    parm excel_definition_file:
        The JSON file which contains the input Excel file's structure.

    """
    excel_definition, excel_columns, stories = __pre_parse_excel_file(
        input_file, excel_definition_file, sprint_schedule_file
    )

    cprint(f"There are {excel_definition.column_count} columns in the definition file.")
    if not excel_definition.count_of_pre_process_steps():
        cprint("No pre-process steps have been configured.", color="light_red")
    else:
        cprint("Pre-process steps:")
        for index, step in enumerate(excel_definition.get_pre_process_steps()):
            cprint(f"{index + 1}: {step.name}. Enabled: {step.enabled}.")

    if not excel_definition.count_of_sort_strategies():
        cprint("No sort strategies have been configured.", color="yellow")
    else:
        cprint("Sort strategies:")
        for index, strategy in enumerate(excel_definition.get_sort_strategies()):
            cprint(f"{index + 1}: {strategy.name}. Enabled: {strategy.enabled}.")

    if stories is None or excel_columns is None:
        cprint("Parse excel file failed.", color="light_red")
        return

    cprint(f"There are {len(excel_columns)} columns in the excel.")

    if not stories:
        cprint("There are no stories inside the excel file.", color="light_yellow")
    else:
        story_need_sort: int = 0
        for story in stories:
            if story.need_sort:
                story_need_sort += 1

        cprint(f"There are {story_need_sort} stories can be sorted.")



# -*- coding: utf-8 -*-
"""
This module offers a set of operations that user can modify their Excel files.
"""
import pathlib
import warnings
from os import remove
from pathlib import Path
from typing import List, Optional, Tuple, TypedDict, Union

import openpyxl
from openpyxl.cell.cell import Cell, MergedCell
from openpyxl.cell.read_only import EmptyCell
from openpyxl.workbook import Workbook
from openpyxl.worksheet._read_only import ReadOnlyWorksheet
from openpyxl.worksheet.worksheet import Worksheet
from urllib3 import disable_warnings

from .excel_definition import ExcelDefinition, ExcelDefinitionColumn
from .sprint_schedule import SprintScheduleStore
from .story import Story, StoryFactory
from .utils import is_absolute_path_valid, standardize_column_name

__all__ = ["read_excel_file", "output_to_excel_file"]

# Currently, the openpyxl package will report an obsolete warning.
warnings.simplefilter(action="ignore", category=UserWarning)
# Disable the HTTPS certificate verification warning.
disable_warnings()


def read_excel_file(
    file: Union[str, Path],
    excel_definition: ExcelDefinition,
    sprint_schedule: SprintScheduleStore,
) -> Tuple[List[str], List[Story]]:
    """
    Read and parse the Excel file

    parm file:
        The Excel file that you want to read

    parm excel_definition:
        The Excel column definition which is imported
        from the :py:class:`ExcelDefinition`

    parm sprint_schedule:
        The priority mapping for the :py:class:`Milestone` object.

    :return:
        A :py:class:`tuple` object which contains a list of column
        name and a list of :py:class:`Story`.
    """
    work_book: Optional[Workbook] = None
    try:
        if not is_absolute_path_valid(file):
            raise FileNotFoundError(
                f"""Please make sure the input excel file exist and
                the path should be absolute. File: {file}."""
            )

        with open(str(file), mode="rb") as raw_file:
            work_book = openpyxl.load_workbook(
                raw_file,
                read_only=True,
                keep_vba=False,
                data_only=True,
                keep_links=True,
            )

            if work_book.active is None or (
                not isinstance(work_book.active, Worksheet)
                and not isinstance(work_book.active, ReadOnlyWorksheet)
            ):
                raise ValueError("The input excel file doesn't contain any sheets.")

            sheet: Union[Worksheet, ReadOnlyWorksheet] = work_book.active

            column_count = sheet.max_column
            if column_count is None:
                column_count = excel_definition.max_column_index

            if sheet.max_row is not None and sheet.max_row < 2:
                return [], []

            actual_excel_columns: List[str] = []
            actual_column_names: List[str] = []
            for column_index in range(1, column_count + 1):
                column_value = str(
                    sheet.cell(row=1, column=column_index).value  # type: ignore[misc]
                )
                if column_value is None or not str(column_value):
                    raise ValueError("The input excel file has invalid/empty column.")
                actual_excel_columns.append(column_value)
                if standardize_column_name(column_value) in actual_column_names:
                    raise ValueError(
                        f"""The input excel file has duplicate column.
Column name: {column_value}."""
                    )
                actual_column_names.append(standardize_column_name(column_value))

            stories: List[Story] = []

            defined_excel_columns = excel_definition.get_columns()

            class _DefinedColumn(TypedDict):
                finded: bool
                name: str
                self: ExcelDefinitionColumn

            defined_columns: List[_DefinedColumn] = [
                {
                    "finded": False,
                    "name": standardize_column_name(i["name"]),
                    "self": i,
                }
                for i in defined_excel_columns
            ]
            story_factory = StoryFactory(defined_excel_columns)

            for row in sheet.iter_rows(  # type: ignore[misc]
                min_row=2,
                max_row=sheet.max_row,
                min_col=1,
                max_col=len(actual_excel_columns),
            ):
                if __should_skip(row):
                    continue

                story: Story = story_factory.create_story()
                story.excel_row_index = __extract_row_number(row)
                actual_column_values: List[Cell | MergedCell] = list(row)

                for actual_column_index, actual_column_name in enumerate(
                    actual_column_names
                ):
                    if not actual_column_name:
                        continue
                    finded = False
                    for finded_defined_excel_column in defined_columns:
                        if finded_defined_excel_column[
                            "name"
                        ] == standardize_column_name(actual_column_name):
                            finded_defined_excel_column["finded"] = True
                            story.set_value(
                                finded_defined_excel_column["self"]["type"],
                                actual_column_name,
                                actual_column_values[actual_column_index].value,
                            )
                            finded = True
                            break
                    if not finded:
                        story.set_value(
                            str,
                            actual_column_name,
                            actual_column_values[actual_column_index].value,
                        )
                not_finded_defined_excel_column_names = [
                    i["self"]["name"] for i in defined_columns if i["finded"] is False
                ]
                if not_finded_defined_excel_column_names:
                    missing_msg = "\n".join(
                        [
                            f"{index + 1}.{name}"
                            for index, name in enumerate(
                                not_finded_defined_excel_column_names
                            )
                        ]
                    )
                    raise ValueError(f"Following columns are missing:\n{missing_msg}.")

                for item in defined_columns:
                    item["finded"] = False

                story.calc_sprint_schedule(sprint_schedule)
                stories.append(story)
        raw_file.close()
    finally:
        if work_book:
            work_book.close()
    return actual_excel_columns, stories


def __should_skip(row: Tuple[Cell | MergedCell, ...]) -> bool:
    is_all_cell_empty = True
    for cell in row:
        if (
            cell
            and not isinstance(cell, EmptyCell)
            and cell.value is not None
            and str(cell.value).strip()
        ):
            is_all_cell_empty = False
    return is_all_cell_empty


def __extract_row_number(row: Tuple[Cell | MergedCell, ...]) -> int:
    for cell in row:
        if cell and not isinstance(cell, (EmptyCell, MergedCell)):
            return int(cell.row)
    return 0


def output_to_excel_file(
    file: Union[str, Path],
    stories: "List[Story]",
    excel_column_names: List[str],
    over_write: bool = True,
):
    """
    Generate Excel file

    parm file:
        Output Excel file name including the path

    parm stories:
        A list of :py:class:`Story` which need to be written to the Excel

    parm excel_column_names:
        Excel columns
        the :py:class:`ExcelDefinition`

    parm over_write:
        Whether the exist output file will be over-write.
    """
    if file is None or not pathlib.Path(file).is_absolute():
        raise ValueError("The output file path is invalid.")

    if pathlib.Path(file).exists():
        if over_write is True:
            try:
                remove(file)
            except PermissionError as e:
                raise FileExistsError(
                    f"The exist excel file: {file} cannot be removed. {e.args[0]}"
                ) from e
        else:
            raise FileExistsError(f"The output excel file: {file} is already exist.")

    work_book = openpyxl.Workbook(write_only=False)

    if work_book.active is None or (
        not isinstance(work_book.active, Worksheet)
        and not isinstance(work_book.active, Worksheet)
    ):
        work_book.close()
        raise ValueError("The output excel file cannot be generated.")

    sheet: Worksheet = work_book.active

    for column_index, column_name in enumerate(excel_column_names):
        cell = sheet.cell(row=1, column=column_index + 1)
        # There are three kinds of Cells. Only the Cell has the value attribute.
        if hasattr(cell, "value"):
            setattr(cell, "value", column_name)

    if stories:
        for row_index, story in enumerate(stories):
            for column_index, column_name in enumerate(excel_column_names):
                cell = sheet.cell(row=row_index + 2, column=column_index + 1)
                if hasattr(cell, "value"):
                    setattr(cell, "value", story.format_value(column_name))

    work_book.save(str(file))
    work_book.close()



from jira_assistant.utils import parse_index_range


def test_parse_index_range():
    result = parse_index_range("1-3")
    assert result is not None
    assert len(result) == 3
    assert 1 in result
    assert 2 in result
    assert 3 in result

    result = parse_index_range("1,3, 5")
    assert result is not None
    assert len(result) == 3
    assert 1 in result
    assert 3 in result
    assert 5 in result


def test_parse_index_range_invalid_cases():
    result = parse_index_range("")
    assert result is None

    result = parse_index_range("abc")
    assert result is None



# -*- coding: utf-8 -*-
"""
This module is used to store excel column definition information.
"""
import pathlib
import warnings
from json import loads
from typing import Any, Dict, List, Optional, Tuple, TypedDict, Union

from jira import JIRA, Issue, JIRAError
from termcolor import cprint
from typing_extensions import NotRequired, Required, Self
from urllib3 import disable_warnings

from .utils import dict_has_key, strip_lower

# Currently, the openpyxl package will report an obsolete warning.
warnings.simplefilter(action="ignore", category=UserWarning)
# Disable the HTTPS certificate verification warning.
disable_warnings()

HERE = pathlib.Path(__file__).resolve().parent
ASSETS = HERE / "assets"
DEFAULT_JIRA_FIELD_TYPE_FILE = ASSETS / "jira_field_type.json"


class JiraFieldTypeDefinition:
    def __init__(
        self,
        type_: Optional[str],
        name: Optional[str],
        properties: List[Self],
        is_basic: Optional[bool],
        array_item_type: Optional[str],
    ) -> None:
        self.__type = type_
        self.__name = name
        self.__properties = properties
        self.__is_basic = is_basic
        self.__array_item_type = array_item_type

    @property
    def type_(self) -> Optional[str]:
        return self.__type

    @type_.setter
    def type_(self, value: str):
        self.__type = value

    @property
    def name(self) -> Optional[str]:
        return self.__name

    @name.setter
    def name(self, value: str):
        self.__name = value

    @property
    def properties(self) -> List[Self]:
        return self.__properties

    @properties.setter
    def properties(self, value: Optional[List[Self]]):
        if value is None:
            self.__properties = []
        else:
            self.__properties = value

    @property
    def is_basic(self) -> Optional[bool]:
        return self.__is_basic

    @is_basic.setter
    def is_basic(self, value: Optional[bool]):
        self.__is_basic = value

    @property
    def array_item_type(self) -> Optional[str]:
        return self.__array_item_type

    @array_item_type.setter
    def array_item_type(self, value: Optional[str]):
        self.__array_item_type = value


_jira_field_types: List[JiraFieldTypeDefinition] = []
_current_jira_field_types_file_path: pathlib.Path = DEFAULT_JIRA_FIELD_TYPE_FILE


def __init_jira_field_types(jira_field_type_file: Optional[pathlib.Path]):
    if not _jira_field_types or (
        jira_field_type_file is not None
        and not _current_jira_field_types_file_path.samefile(jira_field_type_file)
    ):
        if jira_field_type_file is None:
            jira_field_type_file = DEFAULT_JIRA_FIELD_TYPE_FILE
        for i in loads(s=jira_field_type_file.read_text(encoding="utf-8")):
            _jira_field_types.append(__parse_json_to_jira_field_type_definition(i))


def __parse_json_to_jira_field_type_definition(raw: Any) -> JiraFieldTypeDefinition:
    definition = JiraFieldTypeDefinition(
        type_=None,
        is_basic=None,
        properties=[],
        name=None,
        array_item_type=None,
    )
    definition.type_ = raw.get("type", None)
    definition.is_basic = raw.get("isBasic", None)
    definition.name = raw.get("name", None)
    definition.array_item_type = raw.get("arrayItemType", None)
    for property_ in raw.get("properties", []):
        definition.properties.append(
            __parse_json_to_jira_field_type_definition(property_)
        )

    return definition


def get_jira_field(
    field_type: Optional[str], jira_field_type_file: Optional[pathlib.Path] = None
) -> Optional[JiraFieldTypeDefinition]:
    if field_type is None or len(field_type.strip()) == 0:
        return None
    __init_jira_field_types(jira_field_type_file)
    for jira_field_type in _jira_field_types:
        if (
            jira_field_type.type_ is not None
            and jira_field_type.type_.lower() == field_type.lower()
        ):
            return jira_field_type
    return None


class JiraFieldPropertyPathDefinition(TypedDict):
    path: Required[str]
    is_array: Required[bool]


def get_field_paths_of_jira_field(
    field_type: str,
    field_property_name: str,
    jira_field_type_file: Optional[pathlib.Path] = None,
) -> Optional[List[JiraFieldPropertyPathDefinition]]:
    jira_field = get_jira_field(field_type, jira_field_type_file)
    if jira_field is None:
        return None
    if jira_field.is_basic is True:
        return [
            JiraFieldPropertyPathDefinition(path=field_property_name, is_array=False)
        ]
    result: List[JiraFieldPropertyPathDefinition] = []
    is_array_item = jira_field.array_item_type is not None
    # Following code will use the same jira field type file, so no need to pass.
    __internal_get_field_paths_of_jira_field(
        jira_field,
        is_array_item,
        [
            JiraFieldPropertyPathDefinition(
                path=field_property_name,
                is_array=is_array_item,
            )
        ],
        result,
    )
    return result


def __internal_get_field_paths_of_jira_field(
    jira_field: Optional[JiraFieldTypeDefinition],
    is_array_item: bool,
    temp: List[JiraFieldPropertyPathDefinition],
    final: List[JiraFieldPropertyPathDefinition],
):
    if jira_field is None:
        return None
    if jira_field.is_basic is True:
        for item in temp:
            final.append(
                JiraFieldPropertyPathDefinition(
                    path=connect_jira_field_path(item["path"], jira_field.name),
                    is_array=is_array_item,
                )
            )
        temp.clear()
    if jira_field.array_item_type is not None:
        __internal_get_field_paths_of_jira_field(
            get_jira_field(jira_field.array_item_type), True, temp, final
        )
    if jira_field.properties is not None:
        for field_property in jira_field.properties:
            if field_property.array_item_type is not None:
                for item in temp:
                    item["path"] = connect_jira_field_path(
                        item["path"], field_property.name
                    )
                __internal_get_field_paths_of_jira_field(
                    get_jira_field(field_property.array_item_type),
                    True,
                    temp,
                    final,
                )
            if field_property.type_ is None:
                continue
            child_field = get_jira_field(field_property.type_)
            if child_field is None:
                continue
            if child_field.is_basic:
                for item in temp:
                    final.append(
                        JiraFieldPropertyPathDefinition(
                            path=connect_jira_field_path(
                                item["path"], field_property.name
                            ),
                            is_array=is_array_item,
                        )
                    )
            else:
                __internal_get_field_paths_of_jira_field(
                    child_field,
                    is_array_item,
                    [
                        JiraFieldPropertyPathDefinition(
                            path=connect_jira_field_path(
                                item["path"], field_property.name
                            ),
                            is_array=item["is_array"],
                        )
                        for item in temp
                    ],
                    final,
                )
    return None


def connect_jira_field_path(
    *paths: Optional[str], joint_char: str = ".", end_char: str = ""
) -> str:
    return f"{joint_char.join([path for path in paths if path])}{end_char}"


class JiraProject:
    def __init__(self, id_: int, name: str) -> None:
        self.__id = id_
        self.__name = name

    @property
    def id_(self) -> int:
        return self.__id

    @id_.setter
    def id_(self, value: int):
        self.__id = value

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, value: str):
        self.__name = value


class JiraIssueType:
    def __init__(self, id_: int, name: str, project_id: int) -> None:
        self.__id = id_
        self.__name = name
        self.__project_id = project_id

    @property
    def id_(self) -> int:
        return self.__id

    @id_.setter
    def id_(self, value: int):
        self.__id = value

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, value: str):
        self.__name = value

    @property
    def project_id(self) -> int:
        return self.__project_id

    @project_id.setter
    def project_id(self, value: int):
        self.__project_id = value


class JiraField:
    def __init__(
        self,
        required: bool,
        is_array: bool,
        name: str,
        id_: str,
        allowed_values: Optional[Dict[str, List[str]]],
    ) -> None:
        self.__required = required
        self.__is_array = is_array
        self.__name = name
        self.__id = id_
        if allowed_values is not None:
            self.__allowed_values = allowed_values
        else:
            self.__allowed_values = {}

    @property
    def required(self) -> bool:
        return self.__required

    @required.setter
    def required(self, value: bool):
        self.__required = value

    @property
    def is_array(self) -> bool:
        return self.__is_array

    @is_array.setter
    def is_array(self, value: bool):
        self.__is_array = value

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, value: str):
        self.__name = value

    @property
    def id_(self) -> str:
        return self.__id

    @id_.setter
    def id_(self, value: str):
        self.__id = value

    @property
    def allowed_values(self) -> Dict[str, List[str]]:
        return self.__allowed_values

    def is_value_allowed(self, value: Optional[str], jira_field_path: str) -> bool:
        if (
            self.__allowed_values.get(jira_field_path, None) is not None
            and value not in self.__allowed_values[jira_field_path]
        ):
            return False
        return True


_DEFAULT_JIRA_TIMEOUT = 60.0


# Jira are case-sensitive APIs.
class JiraClient:
    def __init__(
        self,
        url: str,
        access_token: str,
        user_email: Optional[str] = None,
        timeout: Optional[float] = None,
    ) -> None:
        self.__is_jira_cloud = is_jira_cloud_url(url)
        # Authentication/Authorization
        # https://developer.atlassian.com/cloud/jira/software/basic-auth-for-rest-apis/
        if timeout is None:
            timeout = _DEFAULT_JIRA_TIMEOUT

        if not self.__is_jira_cloud:
            self.jira = JIRA(
                server=url,
                token_auth=access_token,
                timeout=timeout,
                options={"verify": False},
            )
        elif user_email is not None and access_token is not None:
            self.jira = JIRA(
                server=url,
                basic_auth=(user_email, access_token),
                timeout=timeout,
                options={"verify": False},
            )

        # TODO: Fix this hack later.
        self.jira.deploymentType = ()
        self.__field_cache: Dict[
            str, Dict[str, Optional[List[JiraFieldPropertyPathDefinition]]]
        ] = {}
        self.__project_map: Dict[str, JiraProject] = {}
        # The dict key is project_name
        self.__project_issue_map_using_name: Dict[str, List[JiraIssueType]] = {}
        # The dict key is project_id
        self.__project_issue_map_using_id: Dict[int, List[JiraIssueType]] = {}
        # The dict key is: (project_id, issue_id).
        self.__project_issue_field_map: Dict[Tuple[int, int], List[JiraField]] = {}

    def health_check(self) -> bool:
        try:
            if self.jira.myself() is not None:
                return True
            return False
        except JIRAError:
            return False

    def create_story(self, fields: Dict[str, Any]) -> "Optional[Issue]":
        try:
            create_issue_body = convert_fields_to_create_issue_body(fields)
            return self.jira.create_issue(
                fields=create_issue_body,
                prefetch=False,
            )
        except JIRAError as e:
            cprint(
                f"Calling create story API failed. {self.__extract_error_message(e)}",
                color="light_yellow",
            )
        return None

    def get_jira_browser_link(self, key: str) -> "str":
        return f"{self.jira.server_url}/browse/{key}"

    def get_project_by_project_name(self, project_name: str) -> "Optional[JiraProject]":
        project_name = strip_lower(project_name)
        result = self.__project_map.get(project_name, None)
        if result is None:
            self.get_projects()
        return self.__project_map.get(project_name, None)

    def get_projects(self, force_refresh: bool = False) -> "List[JiraProject]":
        # Otherwise, if there is no project,
        # still will call API to retrieve project list.
        if self.__project_map and force_refresh is not True:
            return list(self.__project_map.values())
        self.__project_map.clear()
        project_response = self.jira.projects()
        for proj in project_response:
            proj_name: Optional[str] = getattr(proj, "key", None)
            proj_id: Optional[int] = getattr(proj, "id", None)
            if proj_name and proj_id:
                self.__project_map[strip_lower(proj_name)] = JiraProject(
                    proj_id, proj_name
                )

                try:
                    issue_types: List[JiraIssueType] = []

                    if self.__is_jira_cloud:
                        get_project_response = self.jira.project(str(proj_id))
                        issue_types.extend(
                            [
                                JiraIssueType(issue_type.id, issue_type.name, proj_id)
                                for issue_type in get_project_response.issueTypes
                                if hasattr(issue_type, "id")
                                and hasattr(issue_type, "name")
                            ]
                        )
                    else:
                        # loading project related issue types
                        response = self.jira.project_issue_types(str(proj_id))
                        issue_types.extend(
                            [
                                JiraIssueType(issue_type.id, issue_type.name, proj_id)
                                for issue_type in response
                                if hasattr(issue_type, "id")
                                and hasattr(issue_type, "name")
                            ]
                        )
                    self.__project_issue_map_using_name[strip_lower(proj_name)] = (
                        issue_types
                    )
                    self.__project_issue_map_using_id[proj_id] = issue_types
                except JIRAError as e:
                    cprint(
                        f"""Get issue types failed. Project: {proj_name}. {self.__extract_error_message(e)}""",  # pylint: disable=line-too-long
                        color="light_yellow",
                    )
                    continue
        return list(self.__project_map.values())

    def get_issue_type_by_project_id_and_issue_name(
        self, project_id: int, issue_name: str
    ) -> "Optional[JiraIssueType]":
        match_result = [
            i
            for i in self.__project_issue_map_using_id[project_id]
            if strip_lower(i.name) == strip_lower(issue_name)
        ]
        if match_result:
            return match_result[0]
        return None

    def get_issue_type_by_project_name_and_issue_name(
        self, project_name: str, issue_name: str
    ) -> "Optional[JiraIssueType]":
        project_name = strip_lower(project_name)
        if project_name not in self.__project_issue_map_using_name:
            return None
        match_result = [
            i
            for i in self.__project_issue_map_using_name[project_name]
            if strip_lower(i.name) == strip_lower(issue_name)
        ]
        if match_result:
            return match_result[0]
        return None

    def get_issue_types(self, project_name: str) -> "List[JiraIssueType]":
        project_name = strip_lower(project_name)
        result = self.__project_issue_map_using_name.get(project_name, [])
        if not result:
            self.get_project_by_project_name(project_name)
        return self.__project_issue_map_using_name.get(project_name, [])

    def get_fields_by_project_id_and_issue_id(
        self, project_id: int, issue_id: int
    ) -> "Tuple[List[JiraField], List[JiraField]]":
        result = self.__project_issue_field_map.get(
            (project_id, issue_id),
            [],
        )
        if not result:
            issue_type_ids: List[int] = []

            if self.__is_jira_cloud:
                project_response = self.jira.project(str(project_id))
                issue_type_ids.extend(
                    [
                        issue_type.id
                        for issue_type in project_response.issueTypes
                        if hasattr(issue_type, "id")
                    ]
                )
            else:
                issue_response = self.jira.project_issue_types(str(project_id))
                issue_type_ids.extend(
                    [
                        issue_type.id
                        for issue_type in issue_response
                        if hasattr(issue_type, "id")
                    ]
                )
            # Loading all issue types and related fields.
            for issue_type_id in issue_type_ids:
                # Should be same as issue_name
                if (
                    issue_type_id is not None
                    and (
                        project_id,
                        issue_type_id,
                    )
                    not in self.__project_issue_field_map
                ):
                    if self.__is_jira_cloud:
                        field_types_cloud = self.jira.createmeta_fieldtypes(
                            str(project_id), str(issue_type_id)
                        )
                        if "fields" in field_types_cloud:
                            self.__project_issue_field_map[
                                (project_id, issue_type_id)
                            ] = [
                                self.__convert_field_type_to_jira_field(field_type)
                                for field_type in field_types_cloud["fields"]
                            ]
                    else:
                        field_types = self.jira.project_issue_fields(
                            str(project_id), str(issue_type_id)
                        )
                        if field_types:
                            self.__project_issue_field_map[
                                (project_id, issue_type_id)
                            ] = [
                                self.__convert_field_type_to_jira_field(field_type.raw)
                                for field_type in field_types
                            ]
        # Try to search again.
        result = self.__project_issue_field_map.get(
            (project_id, issue_id),
            [],
        )
        required_fields = []
        not_required_fields = []
        for item in result:
            if item.required is True:
                required_fields.append(item)
            else:
                not_required_fields.append(item)
        return required_fields, not_required_fields

    @staticmethod
    def __convert_field_type_to_jira_field(field_type: Any) -> "JiraField":
        result: JiraField

        result = JiraField(
            field_type["required"],
            field_type["schema"]["type"] == "array",
            field_type.get("name", ""),
            field_type.get("fieldId", ""),
            None,
        )

        schema: Dict = field_type.get("schema")
        is_array: bool = "items" in schema
        if is_array:
            value_type: str = schema.get("items", "")
        else:
            value_type = schema.get("type", "")

        def __extract_allowed_values(item: Union[str, Dict], pre_key: str):
            if isinstance(item, str):
                if dict_has_key(result.allowed_values, pre_key):
                    result.allowed_values[pre_key].append(item)
                else:
                    result.allowed_values[pre_key] = [item]
            if isinstance(item, dict):
                for _key, _value in item.items():
                    __extract_allowed_values(
                        _value, connect_jira_field_path(pre_key, _key)
                    )

        if "allowedValues" in field_type:
            for allowed_value in field_type.get("allowedValues", []):
                disabled = allowed_value.get("disabled", False)
                if not disabled:
                    for key, value in allowed_value.items():
                        __extract_allowed_values(
                            value, connect_jira_field_path(value_type, key)
                        )

        return result

    def get_all_fields(
        self,
    ) -> "Dict[str, Dict[str, Optional[List[JiraFieldPropertyPathDefinition]]]]":
        if not self.__field_cache:
            for field in self.jira.fields():
                if "schema" not in field.keys():
                    continue

                temp: Dict[str, Optional[List[JiraFieldPropertyPathDefinition]]] = {
                    "id": field["id"],
                }

                class _FieldSchema(TypedDict):
                    type: str
                    items: NotRequired[str]
                    custom: NotRequired[str]
                    customId: NotRequired[int]
                    system: NotRequired[str]

                schema: _FieldSchema = field["schema"]
                property_name = field["id"]
                is_array = "items" in schema
                if is_array:
                    field_type = schema.get("items", None)
                else:
                    field_type = schema.get("type", None)

                if field_type is not None:
                    temp["properties"] = get_field_paths_of_jira_field(
                        field_type, property_name
                    )

                    self.__field_cache[field["name"]] = temp
        return self.__field_cache

    def get_stories_detail(
        self, story_ids: List[str], jira_fields: List[Dict[str, str]]
    ) -> "Dict[str, Dict[str, str]]":
        final_result = {}
        batch_size = 200

        if len(story_ids) > batch_size:
            start_index = 0
            end_index = batch_size
            while end_index <= len(story_ids) and start_index < len(story_ids):
                final_result.update(
                    self.__internal_get_stories_detail(
                        story_ids[start_index:end_index], jira_fields
                    )
                )
                start_index = end_index
                if start_index + batch_size < len(story_ids):
                    end_index = start_index + batch_size
                else:
                    end_index = start_index + (len(story_ids) - end_index)
            return final_result
        return self.__internal_get_stories_detail(story_ids, jira_fields)

    def __internal_get_stories_detail(
        self, story_ids: List[str], jira_fields: List[Dict[str, str]]
    ) -> "Dict[str, Dict[str, str]]":
        id_query = ",".join(
            [f"'{str(story_id).strip()}'" for story_id in story_ids if story_id]
        )

        try:
            search_result: Dict[str, Any] = self.jira.search_issues(
                jql_str=f"id in ({id_query})",
                maxResults=len(story_ids),
                fields=[field["jira_path"].split(".")[0] for field in jira_fields],
                json_result=True,
            )  # type: ignore

            final_result = {}
            for issue in search_result["issues"]:
                fields_result = {}
                for field in jira_fields:
                    # First element in the tuple is jira
                    # field name like "customfield_13210 or status..."
                    field_path = field["jira_path"]
                    # Remain elements represent the property path.
                    # Maybe no fields.
                    if "fields" in issue:
                        field_value: Any = issue["fields"]
                        for field_path_ in field["jira_path"].split("."):
                            if field_value is None:
                                field_value = ""
                                break
                            field_value = field_value.get(field_path_, None)
                        fields_result[field_path] = field_value
                final_result[issue["key"].lower()] = fields_result

            return final_result
        except JIRAError as e:
            cprint(
                f"Calling search API failed. {self.__extract_error_message(e)}",
                color="light_yellow",
            )
        return {}

    @staticmethod
    def __extract_error_message(error: JIRAError) -> "str":
        if hasattr(error, "text") and error.text:
            return error.text

        if (
            error.status_code == 400
            and error.response is not None
            and error.response.text
        ):
            error_response: Dict[str, Any] = error.response.json()
            error_messages = error_response.get("errorMessages", [])
            if error_messages:
                return "|".join(error_messages)
            errors: Dict[str, Any] = error_response.get("errors", {})
            if errors:
                return "|".join([f"{k}: {v}" for k, v in errors.items()])
        return str(error.response.text)


def convert_fields_to_create_issue_body(fields: Dict[str, Any]) -> "Dict[str, Any]":
    issue_fields: Dict[str, Any] = {}
    for key, value in fields.items():
        field_paths = key.split(".")
        tmp = issue_fields
        is_array = isinstance(value, list)
        for count, field_path in enumerate(field_paths):
            # if this value is an array and at least has 2 levels
            # then the last property will be an array.
            if is_array and count == len(field_paths) - 2:
                tmp[field_path] = [
                    {field_paths[len(field_paths) - 1]: v} for v in value
                ]
                break
            if count == len(field_paths) - 1:
                tmp[field_path] = value
            else:
                if tmp.get(field_path, None) is not None:
                    # merge exist dict keys.
                    tmp[field_path] = {**{}, **tmp[field_path]}
                else:
                    tmp[field_path] = {}
            tmp = tmp[field_path]
    return issue_fields


def is_jira_cloud_url(jira_url: Optional[str]) -> bool:
    if jira_url is not None and "atlassian.net".upper() in jira_url.upper():
        return True
    return False



# -*- coding: utf-8 -*-
"""
This module is used to list all exported classes/methods.
"""

from .assistant import (
    run_steps_and_sort_excel_file,
    dry_run_steps_and_sort_excel_file,
    generate_jira_field_mapping_file,
)
from .console_script import (
    generate_template,
    process_excel_file,
    update_jira_info,
    get_package_version,
)
from .excel_definition import ExcelDefinition, ExcelDefinitionColumn
from .excel_operation import output_to_excel_file, read_excel_file
from .milestone import Milestone
from .priority import Priority
from .sprint_schedule import SprintScheduleStore
from .story import Story, StoryFactory

__version__ = get_package_version()

__all__ = [
    "ExcelDefinition",
    "ExcelDefinitionColumn",
    "read_excel_file",
    "output_to_excel_file",
    "run_steps_and_sort_excel_file",
    "dry_run_steps_and_sort_excel_file",
    "generate_jira_field_mapping_file",
    "Milestone",
    "Priority",
    "SprintScheduleStore",
    "Story",
    "StoryFactory",
    "process_excel_file",
    "generate_template",
    "update_jira_info",
]




# -*- coding: utf-8 -*-
from pytest import raises

from jira_assistant.milestone import Milestone
from jira_assistant.sprint_schedule import SprintScheduleStore

from . import SRC_ASSETS


def test_init():
    schedule_filename = SRC_ASSETS / "sprint_schedule.json"
    store = SprintScheduleStore()
    store.load_file(schedule_filename)

    milestone = Milestone("R139")
    milestone.calc_priority(store)
    assert milestone.priority == 1


def test_compare():
    schedule_filename = SRC_ASSETS / "sprint_schedule.json"
    store = SprintScheduleStore()
    store.load_file(schedule_filename)

    m_1 = Milestone("M123")
    m_1.calc_priority(store)
    m_2 = Milestone("M125")
    m_2.calc_priority(store)
    m_3 = Milestone("m125")
    m_3.calc_priority(store)
    m_4 = Milestone("R141")
    m_4.calc_priority(store)
    assert m_1 < m_2
    with raises(AssertionError):
        assert m_1 > m_2
    assert m_1 <= m_2
    with raises(AssertionError):
        assert m_1 >= m_2
    assert m_2 >= m_1
    assert m_2 == m_3
    assert m_3 == m_4
    assert m_4 > m_1
    with raises(AssertionError):
        assert m_1 == m_4


def test_str():
    m_1 = Milestone(None)

    assert len(str(m_1)) == 0



# -*- coding: utf-8 -*-
from jira_assistant.priority import Priority, convert_to_priority


def test_compare_priority():
    p_1: Priority = Priority.CRITICAL
    p_2: Priority = Priority.HIGH
    p_3: Priority = Priority.HIGH
    p_4: Priority = Priority.MIDDLE
    p_5: Priority = Priority.LOW
    p_6: Priority = Priority.NA

    assert p_1 > p_2
    assert p_2 < p_1
    assert p_2 == p_3
    assert p_3 > p_4
    assert p_4 < p_3
    assert p_4 > p_5
    assert p_5 < p_4
    assert p_5 > p_6
    assert p_5 >= p_6
    assert p_6 < p_5
    assert p_6 <= p_5


def test_priority_to_str():
    p_1: Priority = Priority.CRITICAL
    p_2: Priority = Priority.NA
    assert str(p_1) == "Critical"
    assert str(p_2) == "N/A"


def test_convert_to_priority():
    assert convert_to_priority("CRITICAL") == Priority.CRITICAL
    assert convert_to_priority("LOW") == Priority.LOW
    assert convert_to_priority("MIDDLE") == Priority.MIDDLE
    assert convert_to_priority("HIGH") == Priority.HIGH
    assert convert_to_priority(None) == Priority.NA



# -*- coding: utf-8 -*-
from __future__ import annotations

from requests_mock import Mocker

from jira_assistant.jira_client import (
    JiraClient,
    JiraIssueType,
    JiraProject,
    connect_jira_field_path,
    convert_fields_to_create_issue_body,
    get_field_paths_of_jira_field,
    get_jira_field,
    is_jira_cloud_url,
)
from tests.mock_server import (
    mock_jira_requests,
    mock_jira_requests_with_error_response,
    mock_jira_requests_with_failed_status_code,
    mock_jira_stories,
)

from . import ASSETS_FILES

DEFAULT_JIRA_URL = "http://localhost"
DEFAULT_JIRA_CLOUD_URL = "http://jira.atlassian.net"
DEFAULT_JIRA_ACCESS_TOKEN = "123"
DEFAULT_JIRA_USER_EMAIL = "sharry.xu@outlook.com"


def test_is_jira_cloud_url():
    assert is_jira_cloud_url("https://jira.atlassian.net")
    assert not is_jira_cloud_url("https://www.google.com")
    assert not is_jira_cloud_url(None)


def test_jira_project():
    proj = JiraProject(1, "a")
    proj.id_ = 2
    assert proj.id_ == 2
    proj.name = "abc"
    assert proj.name == "abc"


def test_jira_issue_type():
    issue_type = JiraIssueType(1, "a", 1)
    issue_type.id_ = 2
    assert issue_type.id_ == 2
    issue_type.name = "abc"
    assert issue_type.name == "abc"
    issue_type.project_id = 3
    assert issue_type.project_id == 3


def test_connect_jira_field_path():
    assert connect_jira_field_path("a", "b") == "a.b"
    assert connect_jira_field_path("a", None) == "a"
    assert connect_jira_field_path(None, "b") == "b"
    assert connect_jira_field_path(None, None) == ""
    assert connect_jira_field_path("a", "b", "c") == "a.b.c"
    assert connect_jira_field_path("a", "b", "c", joint_char="|") == "a|b|c"
    assert (
        connect_jira_field_path("a", "b", "c", joint_char="|", end_char="&") == "a|b|c&"
    )
    assert connect_jira_field_path() == ""
    assert connect_jira_field_path(end_char="&") == "&"


def test_get_stories_detail():
    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        stories = client.get_stories_detail(
            ["A-1", "A-2", "B-1"],
            [
                {
                    "name": "domain",
                    "jira_name": "customfield_15601",
                    "jira_path": "customfield_15601.value",
                },
                {
                    "name": "status",
                    "jira_name": "status",
                    "jira_path": "status.name",
                },
            ],
        )
        assert len(stories) == 3


def test_get_stories_detail_with_large_amount_of_stories():
    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        stories = client.get_stories_detail(
            list(mock_jira_stories.keys()),
            [
                {
                    "name": "domain",
                    "jira_name": "customfield_15601",
                    "jira_path": "customfield_15601.value",
                },
                {
                    "name": "status",
                    "jira_name": "status",
                    "jira_path": "status.name",
                },
            ],
        )
        assert len(stories) == 247


def test_health_check_for_self_host_jira():
    with Mocker(real_http=False, adapter=mock_jira_requests()):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        assert client.health_check() is True


def test_health_check_for_cloud_jira():
    with Mocker(real_http=False, adapter=mock_jira_requests()):
        client = JiraClient(
            DEFAULT_JIRA_CLOUD_URL,
            DEFAULT_JIRA_ACCESS_TOKEN,
            user_email=DEFAULT_JIRA_USER_EMAIL,
        )

        assert client.health_check() is True


def test_health_check_failed():
    with Mocker(real_http=False, adapter=mock_jira_requests_with_failed_status_code()):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        assert client.health_check() is False


def test_get_stories_detail_failed():
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests_with_error_response(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        stories = client.get_stories_detail(
            ["A-1", "A-2", "B-1"],
            [
                {
                    "name": "domain",
                    "jira_name": "customfield_15601",
                    "jira_path": "customfield_15601.value",
                },
                {
                    "name": "status",
                    "jira_name": "status",
                    "jira_path": "status.name",
                },
            ],
        )
        assert len(stories) == 0


def test_get_stories_detail_with_large_amount_of_stories_failed():
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests_with_error_response(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        stories = client.get_stories_detail(
            list(mock_jira_stories.keys()),
            [
                {
                    "name": "domain",
                    "jira_name": "customfield_15601",
                    "jira_path": "customfield_15601.value",
                },
                {
                    "name": "status",
                    "jira_name": "status",
                    "jira_path": "status.name",
                },
            ],
        )
        assert len(stories) == 0


def test_get_all_fields():
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        result = client.get_all_fields()

        assert len(result) == 5


def test_get_projects():
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        result = client.get_projects()

        result = client.get_projects(force_refresh=True)
        assert len(result) == 3
        assert "POC" in [r.name for r in result]
        assert "SD" in [r.name for r in result]
        assert "APPSEC" in [r.name for r in result]


def test_get_issue_types():
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        result = client.get_issue_types("POC")

        assert len(result) == 7
        assert "Release" in [item.name for item in result]


def test_get_fields_by_project_name_and_issue_name():
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        project = client.get_project_by_project_name("POC")
        assert project is not None

        issue_type = client.get_issue_type_by_project_name_and_issue_name(
            project.name, "Story"
        )
        assert issue_type is not None

        (
            required_fields,
            not_required_fields,
        ) = client.get_fields_by_project_id_and_issue_id(project.id_, issue_type.id_)

        assert len(required_fields) == 3
        assert len(not_required_fields) == 3


def test_get_fields_by_unknown_project_name_and_issue_name():
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        issue_type = client.get_issue_type_by_project_name_and_issue_name(
            "Proj1", "Story1"
        )
        assert issue_type is None


def test_get_fields_by_project_name_and_unknown_issue_name():
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        project = client.get_project_by_project_name("POC")
        assert project is not None

        issue_type = client.get_issue_type_by_project_name_and_issue_name(
            project.name, "Story1"
        )
        assert issue_type is None


def test_get_fields_by_project_id_and_issue_name():
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        project = client.get_project_by_project_name("POC")
        assert project is not None

        issue_type = client.get_issue_type_by_project_id_and_issue_name(
            project.id_, "Story"
        )
        assert issue_type is not None

        (
            required_fields,
            not_required_fields,
        ) = client.get_fields_by_project_id_and_issue_id(project.id_, issue_type.id_)

        assert len(required_fields) == 3
        assert len(not_required_fields) == 3


def test_convert_fields_to_create_issue_body():
    issue_fields = {
        "customfield_11204.value": ["BAU, Mobile"],
        "issuetype.name": "Release Support",
        "issuetype.id": "1000",
        "summary": "test",
        "project.id": 10100,
        "project.key": "Sandbox",
        "customfield_12345": ["ABC"],
        "labels": ["A", "B"],
    }

    create_issue_fields = convert_fields_to_create_issue_body(issue_fields)
    assert isinstance(create_issue_fields["customfield_11204"], list)
    assert create_issue_fields["customfield_11204"][0]["value"] == "BAU, Mobile"
    assert create_issue_fields["issuetype"]["name"] == "Release Support"
    assert create_issue_fields["issuetype"]["id"] == "1000"
    assert create_issue_fields["summary"] == "test"
    assert create_issue_fields["project"]["id"] == 10100
    assert create_issue_fields["project"]["key"] == "Sandbox"
    assert isinstance(create_issue_fields["customfield_12345"], list)
    assert create_issue_fields["customfield_12345"][0] == "ABC"
    assert isinstance(create_issue_fields["labels"], list)
    assert create_issue_fields["labels"][0] == "A"
    assert create_issue_fields["labels"][1] == "B"


def test_get_jira_field_use_basic_type():
    assert get_jira_field(None) is None
    assert get_jira_field("abc") is None
    actual_field = get_jira_field("any")
    assert actual_field is not None
    assert actual_field.is_basic is True


def test_get_jira_field_use_complex_type():
    actual_field = get_jira_field("status")
    assert actual_field is not None
    assert actual_field.is_basic is False
    assert actual_field.type_ == "status"
    assert actual_field.properties is not None
    assert "name" in [p.name for p in actual_field.properties]
    assert "statusCategory" in [p.name for p in actual_field.properties]


def test_get_field_paths_of_jira_field_use_unknown_type():
    actual_field_paths = get_field_paths_of_jira_field(
        "abc", "abc", ASSETS_FILES / "jira_field_type.json"
    )
    assert actual_field_paths is None


def test_get_field_paths_of_jira_field_use_child_array_type():
    actual_field_paths = get_field_paths_of_jira_field(
        "array-level1", "abc", ASSETS_FILES / "jira_field_type.json"
    )
    assert actual_field_paths is not None
    assert "abc.name.level3" in [item["path"] for item in actual_field_paths]


def test_get_field_paths_of_jira_field_use_basic_type():
    actual_field_paths = get_field_paths_of_jira_field("string", "customfield_17001")
    assert actual_field_paths is not None
    assert "customfield_17001" == actual_field_paths[0]["path"]
    assert not actual_field_paths[0]["is_array"]


def test_get_field_paths_of_jira_field_use_complex_type_no_hierarchy():
    # Author
    actual_field_paths = get_field_paths_of_jira_field("author", "abc")
    assert actual_field_paths is not None
    assert len(actual_field_paths) == 2
    assert "abc.name" in [item["path"] for item in actual_field_paths]
    assert "abc.emailAddress" in [item["path"] for item in actual_field_paths]


def test_get_field_paths_of_jira_field_use_complex_type_multiple_hierarchy():
    # Project
    actual_field_paths = get_field_paths_of_jira_field("project", "abc")
    assert actual_field_paths is not None
    assert len(actual_field_paths) == 6
    assert "abc.name" in [item["path"] for item in actual_field_paths]
    assert "abc.key" in [item["path"] for item in actual_field_paths]
    assert "abc.projectTypeKey" in [item["path"] for item in actual_field_paths]
    assert "abc.projectCategory.description" in [
        item["path"] for item in actual_field_paths
    ]
    assert "abc.projectCategory.name" in [item["path"] for item in actual_field_paths]


def test_get_field_paths_of_jira_field_use_array_type_no_hierarchy():
    # comments-page
    actual_field_paths = get_field_paths_of_jira_field("comments-page", "abc")
    assert actual_field_paths is not None
    assert len(actual_field_paths) == 5
    assert "abc.comments.author.name" in [item["path"] for item in actual_field_paths]
    assert "abc.comments.author.emailAddress" in [
        item["path"] for item in actual_field_paths
    ]
    assert "abc.comments.id" in [item["path"] for item in actual_field_paths]
    assert "abc.comments.body" in [item["path"] for item in actual_field_paths]
    assert "abc.comments.created" in [item["path"] for item in actual_field_paths]


def test_create_story_failed(capsys):
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests_with_error_response(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        result = client.create_story(
            {
                "project": {"key": "SD", "id": 10100},
                "summary": "Item 1",
                "description": "Good",
                "issuetype": {"name": "story"},
            }
        )

        assert result is None
        output = capsys.readouterr()
        assert "issuetype: issue type is required" in output.out



# -*- coding: utf-8 -*-
from enum import IntEnum
from typing import Any

__all__ = ["Priority", "convert_to_priority"]


class Priority(IntEnum):
    CRITICAL = 5
    HIGH = 4
    MIDDLE = 3
    LOW = 2
    NA = 1

    def __str__(self) -> str:
        if self is None or self.name == "NA":
            return "N/A"
        return self.name.capitalize()

    def __lt__(self, __o: "object") -> bool:
        left = -1
        if isinstance(self.value, int):
            left = self.value
        right = -1
        if isinstance(__o, int):
            right = __o

        if left < right:
            return True
        return False

    def __gt__(self, __o: "object") -> bool:
        left = -1
        if isinstance(self.value, int):
            left = self.value
        right = -1
        if isinstance(__o, int):
            right = __o

        if left > right:
            return True
        return False

    def __le__(self, __o: "object") -> bool:
        left = -1
        if isinstance(self.value, int):
            left = self.value
        right = -1
        if isinstance(__o, int):
            right = __o

        if left <= right:
            return True
        return False

    def __ge__(self, __o: "object") -> bool:
        left = -1
        if isinstance(self.value, int):
            left = self.value
        right = -1
        if isinstance(__o, int):
            right = __o

        if left >= right:
            return True
        return False

    def __eq__(self, __o: "object") -> bool:
        if self.value == __o:
            return True
        return False


def convert_to_priority(raw: Any) -> Priority:
    if raw is None:
        return Priority.NA
    if isinstance(raw, Priority):
        return raw
    value = str(raw).strip().upper()
    if value in ("N/A", "NA"):
        return Priority.NA
    if value == "LOW":
        return Priority.LOW
    if value in ("MEDIUM", "MIDDLE"):
        return Priority.MIDDLE
    if value == "HIGH":
        return Priority.HIGH
    if value == "CRITICAL":
        return Priority.CRITICAL
    return Priority.NA



# -*- coding: utf-8 -*-
"""
This module is used to provide the console program.
"""
import os
import pathlib
import sys
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
from datetime import datetime
from importlib.metadata import version
from pathlib import Path
from shutil import copyfile
from typing import Optional
from urllib.parse import ParseResult, urlparse

from dotenv import set_key
from termcolor import cprint

from .assistant import (
    dry_run_steps_and_sort_excel_file,
    generate_jira_field_mapping_file,
    run_steps_and_sort_excel_file,
)
from .excel_definition import ExcelDefinition
from .excel_operation import output_to_excel_file

__all__ = [
    "process_excel_file",
    "generate_template",
    "update_jira_info",
    "get_package_version",
]


def get_package_version() -> str:
    return version("jira_assistant")


def get_args_for_process_excel_file() -> Namespace:
    """Process console command's arguments. Command: process_excel_file"""
    parser = ArgumentParser(
        description="Used to pre-process and sort stories",
        formatter_class=ArgumentDefaultsHelpFormatter,
        allow_abbrev=False,
    )

    parser.add_argument("input_file", type=pathlib.Path, help="Source Excel file.")
    parser.add_argument(
        "-o",
        "--output-folder",
        "--output_folder",
        metavar="<path>",
        type=pathlib.Path,
        required=False,
        help="Output folder.",
    )
    parser.add_argument(
        "--excel-definition-file",
        "--excel_definition_file",
        metavar="<path>",
        type=pathlib.Path,
        required=False,
        help="Excel definition file. File format: JSON.",
    )
    parser.add_argument(
        "--sprint-schedule-file",
        "--sprint_schedule_file",
        metavar="<path>",
        type=pathlib.Path,
        required=False,
        help="Milestone priority file. File format: JSON.",
    )
    parser.add_argument(
        "--over-write",
        "--over_write",
        metavar="True|False",
        type=bool,
        required=False,
        default=True,
        help="Whether or not to over write existing file.",
    )
    parser.add_argument(
        "--env-file",
        "--env_file",
        metavar="<path>",
        type=pathlib.Path,
        required=False,
        help="Env file which contains info like jira url.",
    )
    parser.add_argument(
        "--dry-run",
        "--dry_run",
        required=False,
        action="store_true",
        help="Only analyze the input and definition files. No side effects at all.",
    )
    parser.add_argument(
        "--v",
        "--version",
        action="version",
        version=f"%(prog)s {get_package_version()}",
    )

    args = parser.parse_args()

    return args


def process_excel_file() -> None:
    """Console command: process_excel_file"""
    try:
        args = get_args_for_process_excel_file()

        # Pre-Process input file
        input_file_absolute_path: pathlib.Path = (
            pathlib.Path.cwd() / args.input_file.as_posix()
        ).resolve()

        if input_file_absolute_path.suffix.lower() != ".xlsx":
            cprint(
                f"Please provide an Excel file. File: {input_file_absolute_path}.",
                color="light_yellow",
            )
            sys.exit(1)

        if not os.path.exists(input_file_absolute_path):
            cprint(
                f"Input file is not exist. File: {input_file_absolute_path}.",
                color="light_yellow",
            )
            sys.exit(1)

        input_file_name_without_extension = input_file_absolute_path.stem

        # Pre-Process output file
        output_folder_absolute_path: pathlib.Path = (
            input_file_absolute_path.parent.absolute()
        )

        if args.output_folder is not None:
            output_folder_absolute_path = (
                pathlib.Path(args.output_folder).resolve().absolute()
            )

        if not output_folder_absolute_path.exists():
            output_folder_absolute_path.mkdir(parents=True, exist_ok=True)

        output_file_absolute_path: pathlib.Path = (
            output_folder_absolute_path
            / f"{input_file_name_without_extension}_sorted.xlsx"
        ).resolve()

        # Fix duplicate output file issue.
        test_output_file_absolute_path = pathlib.Path(
            output_file_absolute_path
        ).resolve()

        copy_count = 1
        while test_output_file_absolute_path.exists():
            test_output_file_absolute_path = (
                output_folder_absolute_path
                / f"{ output_file_absolute_path.stem }_{copy_count}.xlsx"
            )
            copy_count += 1

        output_file_absolute_path = pathlib.Path(
            test_output_file_absolute_path
        ).resolve()

        excel_definition_file_absolute_path = None

        if args.excel_definition_file is not None:
            excel_definition_file_absolute_path = pathlib.Path(
                pathlib.Path.cwd() / args.excel_definition_file.as_posix()
            ).resolve()

        sprint_schedule_file_absolute_path = None

        if args.sprint_schedule_file is not None:
            sprint_schedule_file_absolute_path = pathlib.Path(
                pathlib.Path.cwd() / args.sprint_schedule_file.as_posix()
            ).resolve()

        # Overwrite parameter.
        over_write = True
        if args.over_write is not None:
            over_write = args.over_write

        # Env file parameter.
        env_file = None
        if args.env_file is not None:
            env_file = args.env_file

        # Dry run
        if args.dry_run is True:
            dry_run_steps_and_sort_excel_file(
                input_file_absolute_path,
                excel_definition_file_absolute_path,
                sprint_schedule_file_absolute_path,
            )
        else:
            run_steps_and_sort_excel_file(
                input_file_absolute_path,
                output_file_absolute_path,
                excel_definition_file_absolute_path,
                sprint_schedule_file_absolute_path,
                over_write,
                env_file,
            )

        sys.exit(0)
    except Exception as e:
        cprint(e, color="light_red")
        sys.exit(1)


def get_args_for_generate_template() -> Namespace:
    """Process console command's arguments. Command: generate_template"""
    parser = ArgumentParser(
        description="Used to generate templates",
        formatter_class=ArgumentDefaultsHelpFormatter,
        allow_abbrev=False,
    )

    parser.add_argument(
        "template_type",
        metavar="template_type",
        type=str,
        help="""What kind of file template you want to generate.
Choices: excel, excel-definition, sprint-schedule or jira-field-mapping.""",
        choices=["excel", "excel-definition", "sprint-schedule", "jira-field-mapping"],
    )
    parser.add_argument(
        "-o",
        "--output-folder",
        "--output_folder",
        metavar="<path>",
        type=pathlib.Path,
        required=False,
        help="Output folder.",
    )
    parser.add_argument(
        "--env-file",
        "--env_file",
        metavar="<path>",
        type=pathlib.Path,
        required=False,
        help="Env file which contains info like jira url.",
    )
    parser.add_argument(
        "--v",
        "--version",
        action="version",
        version=f"%(prog)s {get_package_version()}",
    )

    return parser.parse_args()


HERE = pathlib.Path(__file__).resolve().parent
SRC_ASSETS = HERE / "assets"


def generate_template():
    """Console command: generate_template"""
    try:
        args = get_args_for_generate_template()

        template_type: str = str(args.template_type).lower()
        output_folder = Path.cwd()

        if args.output_folder is not None:
            output_folder = args.output_folder

        result: Optional[Path] = None
        if template_type == "excel":
            result = __generate_excel_template(
                __generate_timestamp_filename(output_folder, "excel-template", ".xlsx")
            )
        elif template_type == "excel-definition":
            result = copyfile(
                SRC_ASSETS / "excel_definition.json",
                __generate_timestamp_filename(
                    output_folder, "excel-definition-template", ".json"
                ),
            )
        elif template_type == "sprint-schedule":
            result = copyfile(
                SRC_ASSETS / "sprint_schedule.json",
                __generate_timestamp_filename(
                    output_folder, "sprint-schedule-template", ".json"
                ),
            )
        elif template_type == "jira-field-mapping":
            result = __generate_jira_field_mapping_template(
                __generate_timestamp_filename(
                    output_folder, "jira-field-mapping", ".json"
                ),
                args.env_file,
            )
        else:
            cprint(
                """Invalid template type.
Choices: excel, excel-definition or sprint-schedule.""",
                color="light_red",
            )

        if result is not None and result.is_file():
            cprint(
                f"Generate success! Template type: {template_type}.",
                color="light_green",
            )
            sys.exit(0)
        else:
            cprint(
                f"Generate failed! Template type: {template_type}.", color="light_red"
            )
            sys.exit(1)
    except Exception as e:
        cprint(e, color="light_red")
        sys.exit(1)


def __generate_timestamp_filename(
    output_folder: Path, prefix: str, extension: str
) -> "Path":
    return (
        output_folder
        / f'{prefix}-{datetime.now().strftime("%y-%m-%d-%H-%M-%S")}{extension}'
    ).resolve()


def __generate_excel_template(output_file: "Path") -> Optional[Path]:
    try:
        excel_definition = ExcelDefinition().load_file(
            SRC_ASSETS / "excel_definition.json"
        )
        output_to_excel_file(
            output_file,
            [],
            excel_column_names=excel_definition.get_columns_name(standardized=False),
        )
        return output_file
    except Exception as e:
        cprint(e, color="light_red")
        return None


def __generate_jira_field_mapping_template(
    output_file: "Path", env_file: Optional[Path] = None
) -> Optional[Path]:
    try:
        if generate_jira_field_mapping_file(output_file, env_file=env_file):
            return output_file
    except Exception as e:
        cprint(e, color="light_red")
    return None


def get_args_for_update_jira_info() -> Namespace:
    """Process console command's arguments. Command: update_jira_info"""
    parser = ArgumentParser(
        description="Used to add/update jira url, access token and user email.",
        formatter_class=ArgumentDefaultsHelpFormatter,
        allow_abbrev=False,
    )

    parser.add_argument(
        "--access-token",
        "--access_token",
        metavar="<token>",
        type=str,
        required=False,
        help="Please follow the documentation to get your own access token.",
    )
    parser.add_argument(
        "--url",
        metavar="<jira url>",
        type=str,
        required=False,
        help="Please provide the JIRA website url.",
    )
    parser.add_argument(
        "--user-email",
        "--user_email",
        metavar="<user email>",
        type=str,
        required=False,
        help="Please provide the email address that you use in the JIRA website.",
    )
    parser.add_argument(
        "--env-file",
        "--env_file",
        metavar="<path>",
        type=pathlib.Path,
        required=False,
        help="Custom env file",
    )
    parser.add_argument(
        "--v",
        "--version",
        action="version",
        version=f"%(prog)s {get_package_version()}",
    )

    return parser.parse_args()


def update_jira_info():
    """Console command: update_jira_info"""
    try:
        args = get_args_for_update_jira_info()

        env_file: Path
        if args.env_file is not None:
            env_file = args.env_file
        else:
            env_file = SRC_ASSETS / ".env"

        if not env_file.exists():
            env_file.touch()

        # URL Part
        if args.url is not None:
            parsed_url: ParseResult = urlparse(str(args.url))

            if parsed_url.scheme not in ("https", "http"):
                cprint("Please check the jira url.", color="light_red")
            else:
                result, _, _ = set_key(
                    env_file,
                    key_to_set="JIRA_URL",
                    value_to_set=f"{parsed_url.scheme}://{parsed_url.netloc}",
                    quote_mode="never",
                )

                if result is True:
                    cprint("Add/Update jira url success!", color="light_green")
                else:
                    cprint("Add/Update jira url failed!", color="light_red")

        # ACCESS TOKEN Part
        if args.access_token is not None:
            access_token: str = str(args.access_token)

            if len(access_token.strip()) == 0 or access_token.isspace():
                cprint("Please check the access token.", color="light_red")
                sys.exit(1)
            else:
                result, _, _ = set_key(
                    env_file,
                    key_to_set="JIRA_ACCESS_TOKEN",
                    value_to_set=access_token,
                    quote_mode="never",
                )

                if result is True:
                    cprint("Add/Update jira access token success!", color="light_green")
                else:
                    cprint("Add/Update jira access token failed!", color="light_red")

        # USER EMAIL Part
        if args.user_email is not None:
            user_email: str = str(args.user_email)

            if len(user_email.strip()) == 0 or user_email.isspace():
                cprint("Please check the email address.", color="light_red")
                sys.exit(1)
            else:
                result, _, _ = set_key(
                    env_file,
                    key_to_set="JIRA_USER_EMAIL",
                    value_to_set=user_email,
                    quote_mode="never",
                )

                if result is True:
                    cprint("Add/Update jira user email success!", color="light_green")
                    sys.exit(0)
                else:
                    cprint("Add/Update jira user email failed!", color="light_red")
                    sys.exit(1)
        sys.exit(0)

    except Exception as e:
        cprint(e, color="light_red")
        sys.exit(1)



# -*- coding: utf-8 -*-
"""
This module is used to store excel column definition information.
"""
# pylint: disable=line-too-long
from __future__ import annotations

from copy import deepcopy
from datetime import datetime
from json import loads
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Any, Dict, List, Optional, TypedDict, Union

from .milestone import Milestone
from .priority import Priority
from .utils import (
    dict_has_key,
    is_absolute_path_valid,
    is_index_range_valid,
    standardize_column_name,
    strip_lower,
)

__all__ = ["ExcelDefinition", "ExcelDefinitionColumn", "SortStrategy"]


class BasicStep:
    def __init__(
        self, name: str, enabled: bool, priority: int, config: Optional[Dict[str, Any]]
    ) -> None:
        self.__name: str = name
        self.__enabled: bool = enabled
        self.__priority: int = priority
        self.__config: Optional[Dict[str, Any]] = config

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, value: str):
        self.__name = value

    @property
    def enabled(self) -> bool:
        return self.__enabled

    @enabled.setter
    def enabled(self, value: bool):
        self.__enabled = value

    @property
    def priority(self) -> int:
        return self.__priority

    @priority.setter
    def priority(self, value: int):
        self.__priority = value

    @property
    def config(self) -> Optional[Dict[str, Any]]:
        return self.__config

    @config.setter
    def config(self, value: Dict[str, Any]):
        self.__config = value

    def get_config(self, name: str) -> Optional[Any]:
        if self.__config is not None:
            for key in self.__config.keys():
                if strip_lower(key) == strip_lower(name):
                    return self.__config[key]
        return None


class PreProcessStep(BasicStep):
    pass


def parse_json_item_to_pre_process_step(json_item: Any) -> PreProcessStep:
    pre_process_step_name = ""
    pre_process_step_enabled = False
    pre_process_step_priority = 0
    pre_process_step_config: Dict[str, Any] = {}

    for key, value in json_item.items():
        if strip_lower(key) == strip_lower("name"):
            if value is None or not isinstance(value, str):
                raise ValueError(
                    "The Name property in the pre-process step is invalid."
                )
            pre_process_step_name = value
        if strip_lower(key) == strip_lower("enabled"):
            if value is None or not isinstance(value, bool):
                pre_process_step_enabled = False
            else:
                pre_process_step_enabled = value
        if strip_lower(key) == strip_lower("priority"):
            if value is None or not isinstance(value, int):
                raise ValueError(
                    "The Priority property in the pre-process step is invalid."
                )
            pre_process_step_priority = value
        if strip_lower(key) == strip_lower("config"):
            pre_process_step_config = {}
            if value is not None and isinstance(value, dict):
                for key_, value_ in value.items():
                    pre_process_step_config[key_] = value_

    return PreProcessStep(
        name=pre_process_step_name,
        enabled=pre_process_step_enabled,
        priority=pre_process_step_priority,
        config=pre_process_step_config,
    )


class SortStrategy(BasicStep):
    pass


def parse_json_item_to_sort_strategy(json_item: Any) -> SortStrategy:
    strategy_name = ""
    strategy_priority = 0
    strategy_enabled = False
    strategy_config: Dict[str, Any] = {}

    for key, value in json_item.items():
        if strip_lower(key) == strip_lower("name"):
            if value is None or not isinstance(value, str):
                raise ValueError("The Name property in the sort strategy is invalid.")
            strategy_name = value
        if strip_lower(key) == strip_lower("enabled"):
            if value is None or not isinstance(value, bool):
                strategy_enabled = False
            else:
                strategy_enabled = value
        if strip_lower(key) == strip_lower("priority"):
            if value is None or not isinstance(value, int):
                raise ValueError(
                    "The Priority property in the sort strategy is invalid."
                )
            strategy_priority = value
        if strip_lower(key) == strip_lower("config"):
            strategy_config = {}
            if value is not None and isinstance(value, dict):
                for key_, value_ in value.items():
                    strategy_config[key_] = value_

    return SortStrategy(
        name=strategy_name,
        priority=strategy_priority,
        enabled=strategy_enabled,
        config=strategy_config,
    )


class ExcelDefinitionColumnJiraFieldMapping(TypedDict):
    name: str
    path: str


class ExcelDefinitionColumn(TypedDict):
    index: int
    name: str
    type: Optional[type]
    require_sort: bool
    sort_order: bool
    scope_require_sort: bool
    scope_sort_order: bool
    inline_weights: int
    raise_ranking: int
    scope_raise_ranking: int
    jira_field_mapping: Optional[ExcelDefinitionColumnJiraFieldMapping]
    query_jira_info: bool
    update_jira_info: bool
    delimiter: str


_DEFAULT_VALUE_DELIMITER = "|"


def parse_json_item_to_excel_definition_column(json_item: Any) -> ExcelDefinitionColumn:
    column_index = 0
    column_name = ""
    column_type = None
    column_require_sort = False
    column_sort_order = False
    column_scope_require_sort = False
    column_scope_sort_order = False
    column_inline_weights = -1
    column_raise_ranking = -1
    column_scope_raise_ranking = -1
    column_jira_field_mapping = None
    column_query_jira_info = False
    column_update_jira_info = False
    column_delimiter = _DEFAULT_VALUE_DELIMITER

    for key, value in json_item.items():
        if strip_lower(key) == strip_lower("index"):
            if value is None:
                raise ValueError("Column definition must has an index.")
            if isinstance(value, int):
                column_index = value
            else:
                raise TypeError(
                    "The Index property type in the column definition is not integer."
                )
        elif strip_lower(key) == strip_lower("name"):
            if value is None:
                raise ValueError("Column definition must has a name.")
            if isinstance(value, str):
                column_name = value
            else:
                raise TypeError(
                    "The Name property type in the column definition should be string."
                )
        elif strip_lower(key) == strip_lower("type"):
            column_type = ExcelDefinition.convert_str_to_type(value)
        elif strip_lower(key) == strip_lower("RequireSort"):
            column_require_sort = value
        elif strip_lower(key) == strip_lower("SortOrder"):
            column_sort_order = value
        elif strip_lower(key) == strip_lower("ScopeRequireSort"):
            column_scope_require_sort = value
        elif strip_lower(key) == strip_lower("ScopeSortOrder"):
            column_scope_sort_order = value
        elif strip_lower(key) == strip_lower("InlineWeights"):
            column_inline_weights = value
        elif strip_lower(key) == strip_lower("RaiseRanking"):
            column_raise_ranking = value
        elif strip_lower(key) == strip_lower("ScopeRaiseRanking"):
            column_scope_raise_ranking = value
        elif strip_lower(key) == strip_lower("JiraFieldMapping"):
            column_jira_field_mapping = value
        elif strip_lower(key) == strip_lower("QueryJiraInfo"):
            if value is not None:
                column_query_jira_info = value
        elif strip_lower(key) == strip_lower("UpdateJiraInfo"):
            if value is not None:
                column_update_jira_info = value
        elif strip_lower(key) == strip_lower("Delimiter"):
            if value is not None:
                column_delimiter = strip_lower(value)

    return ExcelDefinitionColumn(
        index=column_index,
        name=column_name,
        type=column_type,
        require_sort=column_require_sort,
        sort_order=column_sort_order,
        scope_require_sort=column_scope_require_sort,
        scope_sort_order=column_scope_sort_order,
        inline_weights=column_inline_weights,
        raise_ranking=column_raise_ranking,
        scope_raise_ranking=column_scope_raise_ranking,
        jira_field_mapping=column_jira_field_mapping,
        query_jira_info=column_query_jira_info,
        update_jira_info=column_update_jira_info,
        delimiter=column_delimiter,
    )


class ExcelDefinition:
    def __init__(self) -> None:
        self.__version = 1
        self.__columns: list[ExcelDefinitionColumn] = []
        self.__sort_strategies: list[SortStrategy] = []
        self.__pre_process_steps: list[PreProcessStep] = []

    def load(self, content: str) -> "ExcelDefinition":
        """
        Load json string to generate the Excel definition

        :param content:
            JSON string content
        """

        try:
            raw_data = loads(s=content)
        except JSONDecodeError as e:
            raise SyntaxError(
                f"The structure of excel definition file is wrong. Hint: {e.msg} in line {e.lineno}:{e.colno}."
            ) from e

        parse_errors = []

        for raw_part in raw_data:
            if not isinstance(raw_part, dict):
                continue
            for name, configuration in raw_part.items():
                if strip_lower(name) == strip_lower("Version") and isinstance(
                    configuration, int
                ):
                    self.__version = configuration
                elif strip_lower(name) == strip_lower("PreProcessSteps"):
                    for item in configuration:
                        try:
                            self.__pre_process_steps.append(
                                parse_json_item_to_pre_process_step(item)
                            )
                        except (TypeError, ValueError) as e:
                            parse_errors.append(e.args[0])
                elif strip_lower(name) == strip_lower("SortStrategies"):
                    for item in configuration:
                        try:
                            self.__sort_strategies.append(
                                parse_json_item_to_sort_strategy(item)
                            )
                        except (TypeError, ValueError) as e:
                            parse_errors.append(e.args[0])
                elif strip_lower(name) == strip_lower("Columns"):
                    for item in configuration:
                        try:
                            self.__columns.append(
                                parse_json_item_to_excel_definition_column(item)
                            )
                        except (TypeError, ValueError) as e:
                            parse_errors.append(e.args[0])

        if parse_errors:
            # Avoid duplicate error messages.
            parse_error_message = "\n".join(
                [f"{index + 1}. {err}" for index, err in enumerate(set(parse_errors))]
            )
            raise SyntaxError(
                f"The excel definition file has below issues need to be fixed:\n{parse_error_message}"
            )

        return self

    def load_file(self, file: Union[str, Path]) -> "ExcelDefinition":
        """
        Load json file to generate the Excel definition

        :param file:
            JSON file location
        """

        if not is_absolute_path_valid(file):
            raise FileNotFoundError(
                f"Please make sure the excel definition file exist and the path should be absolute. File: {file}."
            )

        with open(file=file, mode="r", encoding="utf-8") as table_definition_file:
            try:
                self.load(table_definition_file.read())
            finally:
                table_definition_file.close()

        return self

    def validate(self) -> "List[str]":
        return (
            self.__validate_pre_process_steps()
            + self.__validate_sort_strategies()
            + self.__validate_column_definitions()
        )

    def __validate_pre_process_steps(self) -> "List[str]":
        invalid_definitions: List[str] = []
        valid_pre_process_steps = [
            "CreateJiraStory".lower(),
            "FilterOutStoryWithoutId".lower(),
            "RetrieveJiraInformation".lower(),
            "FilterOutStoryBasedOnJiraStatus".lower(),
        ]

        # Validate PreProcessSteps
        pre_process_step_priorities: List[int] = []
        for pre_process_step in self.__pre_process_steps:
            if pre_process_step.name.lower() not in valid_pre_process_steps:
                invalid_definitions.append("The PreProcessStep name is invalid.")
                continue

            # Validate CreateJiraStory
            if (
                pre_process_step.name.lower() == "CreateJiraStory".lower()
                and pre_process_step.enabled
            ):
                # ProjectType: ProjectIdOrName
                if (
                    standardize_column_name("ProjectType")
                    not in self.get_columns_name()
                ):
                    invalid_definitions.append(
                        "The PreProcessStep: CreateJiraStory must have a column named ProjectType."
                    )
                # IssueType: IssueIdOrName
                if standardize_column_name("IssueType") not in self.get_columns_name():
                    invalid_definitions.append(
                        "The PreProcessStep: CreateJiraStory must have a column named IssueType."
                    )
                continue

            # Validate FilterOutStoryBasedOnJiraStatus
            if (
                pre_process_step.name.lower()
                == "FilterOutStoryBasedOnJiraStatus".lower()
                and pre_process_step.enabled
            ):
                # ProjectType: ProjectIdOrName
                if standardize_column_name("Status") not in self.get_columns_name():
                    invalid_definitions.append(
                        "The PreProcessStep: FilterOutStoryBasedOnJiraStatus must have a column named Status."
                    )

                retrieve_jira_info_step = self.get_pre_process_step_by_name(
                    "RetrieveJiraInformation"
                )
                if (
                    retrieve_jira_info_step is None
                    or retrieve_jira_info_step.priority >= pre_process_step.priority
                ):
                    invalid_definitions.append(
                        "The step named RetrieveJiraInformation must be processed before FilterOutStoryBasedOnJiraStatus."
                    )
                continue

            if (
                pre_process_step.priority is None
                or not isinstance(pre_process_step.priority, int)
                or pre_process_step.priority < 0
            ):
                invalid_definitions.append(
                    f"The pre-process step priority is invalid. PreProcessStep: {pre_process_step.name}"
                )
            else:
                if pre_process_step.priority in pre_process_step_priorities:
                    invalid_definitions.append(
                        f"The pre-process step priority is duplicate. PreProcessStep: {pre_process_step.name}"
                    )
                else:
                    pre_process_step_priorities.append(pre_process_step.priority)

            if pre_process_step.config is not None and dict_has_key(
                pre_process_step.config, "JiraStatuses"
            ):
                if (
                    pre_process_step.name.lower()
                    != "FilterOutStoryBasedOnJiraStatus".lower()
                ):
                    invalid_definitions.append(
                        f"Only FilterOutStoryBasedOnJiraStatus step support JiraStatuses config. PreProcessStep: {pre_process_step.name}."
                    )

                if not isinstance(pre_process_step.config["JiraStatuses"], list):
                    invalid_definitions.append(
                        f"The format of the Jira Statuses is invalid. PreProcessStep: {pre_process_step.name}. Supported format like: ['CLOSED', 'PENDING RELEASE']."
                    )

        return invalid_definitions

    def __validate_sort_strategies(self) -> "List[str]":
        invalid_definitions: List[str] = []

        # Validate Strategies
        strategy_priorities: List[int] = []
        for strategy in self.__sort_strategies:
            if strategy.name.isspace() or len(strategy.name) == 0:
                invalid_definitions.append("The strategy name is invalid.")
                # If strategy name is invalid, no need to check more.
                continue

            if (
                strategy.priority is None
                or not isinstance(strategy.priority, int)
                or strategy.priority < 0
            ):
                invalid_definitions.append(
                    f"The strategy priority is invalid. Strategy: {strategy.name}"
                )
            else:
                if strategy.priority in strategy_priorities:
                    invalid_definitions.append(
                        f"The strategy priority is duplicate. Strategy: {strategy.name}"
                    )
                else:
                    strategy_priorities.append(strategy.priority)

            if strategy.config is not None and dict_has_key(
                strategy.config, "ParentScopeIndexRange"
            ):
                if (
                    strategy.name.lower() != "SortOrder".lower()
                    and strategy.name.lower() != "RaiseRanking".lower()
                ):
                    invalid_definitions.append(
                        f"Only RaiseRanking and SortOrder strategy support ParentScopeIndexRange config. Strategy: {strategy.name}."
                    )
                if not is_index_range_valid(
                    strategy.get_config("ParentScopeIndexRange")
                ):
                    invalid_definitions.append(
                        f"The format of the Parent Level Index Range is invalid. Strategy: {strategy.name}. Supported format strings like: 1-20 or 20,30."
                    )

        return invalid_definitions

    def __validate_column_definitions(  # pylint: disable=too-many-branches
        self,
    ) -> "List[str]":
        invalid_definitions: List[str] = []

        # Validate the Columns
        exist_story_id_column = False
        exist_indexes = []
        exist_column_names = []
        exist_inline_weights = []
        exist_jira_field_names = []
        exist_jira_field_paths = []
        for column in self.get_columns():
            column_index: int = column["index"]
            column_name: str = column["name"]
            column_type: Optional[type] = column["type"]
            column_require_sort: bool = column["require_sort"]
            column_sort_order: bool = column["sort_order"]
            column_scope_require_sort: bool = column["scope_require_sort"]
            column_scope_sort_order: bool = column["scope_sort_order"]
            column_inline_weights: int = column["inline_weights"]
            column_raise_ranking: int = column["raise_ranking"]
            column_scope_raise_ranking: int = column["scope_raise_ranking"]
            column_jira_field_mapping: Optional[
                ExcelDefinitionColumnJiraFieldMapping
            ] = column["jira_field_mapping"]

            # Check Name cannot be empty
            if len(column_name) == 0:
                invalid_definitions.append(
                    f"Column name cannot be empty. Index: {column_index}"
                )
                continue

            if standardize_column_name(column_name) in exist_column_names:
                invalid_definitions.append(
                    f"Column named {column_name} has been duplicated."
                )
                continue
            exist_column_names.append(standardize_column_name(column_name))

            if standardize_column_name(column_name) == standardize_column_name(
                "StoryId"
            ):
                exist_story_id_column = True

            # Check Missing/Duplicate Index
            if not isinstance(column_index, int):
                invalid_definitions.append(
                    f"Column Index can only be number. Column: {column_name}"
                )
            elif column_index is None:
                invalid_definitions.append(f"Missing Index. Column: {column_name}")
            elif column_index in exist_indexes:
                invalid_definitions.append(f"Duplicate Index. Column: {column_name}")
            else:
                exist_indexes.append(column_index)
            # Check Property Type
            if column_type not in (
                str,
                bool,
                datetime,
                Priority,
                Milestone,
                float,
            ):
                invalid_definitions.append(
                    f"Invalid Column Type. Column: {column_name}"
                )

            # Check Sort
            if not isinstance(column_require_sort, bool):
                invalid_definitions.append(
                    f"Require Sort can only be True/False. Column: {column_name}"
                )

            if not isinstance(column_sort_order, bool):
                invalid_definitions.append(
                    f"Sort Order can only be True(Descending)/False(Ascending). Column: {column_name}"
                )

            # Check Sort
            if not isinstance(column_scope_require_sort, bool):
                invalid_definitions.append(
                    f"Scope Require Sort can only be True/False. Column: {column_name}"
                )

            if not isinstance(column_scope_sort_order, bool):
                invalid_definitions.append(
                    f"Scope Sort Order can only be True(Descending)/False(Ascending). Column: {column_name}"
                )

            # Check InlineWeights
            if not isinstance(column_inline_weights, int):
                invalid_definitions.append(
                    f"Inline Weights can only be number. Column: {column_name}"
                )
            else:
                if column_type not in (Priority,) and column_inline_weights > 0:
                    invalid_definitions.append(
                        f"Column do not support Inline Weights. Currently only Priority type support. Column: {column_name}"
                    )
                if (
                    column_inline_weights > 0
                    and column_inline_weights in exist_inline_weights
                ):
                    invalid_definitions.append(
                        f"Duplicate Inline Weights. Currently only support different line weights. Column: {column_name}"
                    )
                exist_inline_weights.append(column_inline_weights)

            # Check RaiseRanking
            if not isinstance(column_raise_ranking, int):
                invalid_definitions.append(
                    f"Raise Ranking can only be number. Column: {column_name}"
                )
            else:
                # Check Support RaiseRanking or not
                if column_type not in (bool,) and column_raise_ranking > 0:
                    invalid_definitions.append(
                        f"Column do not support Raise Ranking feature. Column: {column_name}"
                    )

            if not isinstance(column_scope_raise_ranking, int):
                invalid_definitions.append(
                    f"Scope Raise Ranking can only be number. Column: {column_name}"
                )
            else:
                # Check Support RaiseRanking or not
                if column_type not in (bool,) and column_scope_raise_ranking > 0:
                    invalid_definitions.append(
                        f"Column do not support Scope Raise Ranking feature. Column: {column_name}"
                    )

            if column_jira_field_mapping is None:
                continue
            if column_jira_field_mapping is not None and not isinstance(
                column_jira_field_mapping, dict
            ):
                invalid_definitions.append(
                    f"Jira Field Mapping can only be dictionary. Column: {column_name}"
                )
            else:
                jira_field_name = column_jira_field_mapping.get("name", None)
                if jira_field_name is None or jira_field_name.isspace():
                    invalid_definitions.append(
                        f"Jira Field Mapping has the invalid name. Column: {column_name}"
                    )
                if jira_field_name in exist_jira_field_names:
                    invalid_definitions.append(
                        f"Column has duplicate jira field name. Name: {jira_field_name}, Column: {column_name}"
                    )
                exist_jira_field_names.append(jira_field_name)

                jira_field_path = column_jira_field_mapping.get("path", None)
                if jira_field_path is None or jira_field_path.isspace():
                    invalid_definitions.append(
                        f"Jira Field Mapping has the invalid path. Column: {column_name}"
                    )
                if jira_field_path in exist_jira_field_paths:
                    invalid_definitions.append(
                        f"Column has duplicate jira field path. Path: {jira_field_path}, Column: {column_name}"
                    )
                exist_jira_field_paths.append(jira_field_path)

        # If there is step name 'CreateJiraStory' or 'RetrieveJiraInformation'
        # or 'FilterOutStoryWithoutId', the StoryId must exist.
        if (
            (
                self.get_pre_process_step_by_name("CreateJiraStory")
                or self.get_pre_process_step_by_name("FilterOutStoryWithoutId")
                or self.get_pre_process_step_by_name("RetrieveJiraInformation")
            )
            and len(self.__columns) > 0
            and exist_story_id_column is False
        ):
            invalid_definitions.append(
                "Must have a column named StoryId so that program can identify the record."
            )

        if len(invalid_definitions) == 0:
            self.__columns.sort(key=lambda c: c["index"], reverse=False)

            if len(self.__columns) > 0 and (
                self.__columns[0]["index"] != 1
                or self.__columns[len(self.__columns) - 1]["index"]
                != len(self.__columns)
            ):
                invalid_definitions.append(
                    "Column Index must be in continuation and starts from 1."
                )

        return invalid_definitions

    @staticmethod
    def convert_str_to_type(type_str: str) -> Optional[type]:
        if type_str is None or not isinstance(type_str, str):
            return None
        type_str = strip_lower(type_str)
        if type_str == "str":
            return str
        if type_str == "bool":
            return bool
        if type_str == "datetime":
            return datetime
        if type_str == "priority":
            return Priority
        if type_str == "milestone":
            return Milestone
        # Currently, only support float/double
        if type_str == "number":
            return float
        return None

    def __iter__(self):
        yield from self.__columns

    def get_columns(self) -> "List[ExcelDefinitionColumn]":
        return deepcopy(self.__columns)

    def get_column_by_jira_field_mapping_name(
        self, name: str
    ) -> "List[ExcelDefinitionColumn]":
        result: List[ExcelDefinitionColumn] = []
        for item in self.__columns:
            jira_field_mapping = item.get("jira_field_mapping", None)
            if jira_field_mapping is not None and standardize_column_name(
                jira_field_mapping["path"].split(".")[0]
            ) == standardize_column_name(name):
                result.append(deepcopy(item))
        return result

    def get_columns_name(self, standardized: bool = True) -> "List[str]":
        if standardized:
            return [
                standardize_column_name(item.get("name", ""))
                for item in self.__columns
            ]
        return [item["name"] for item in self.__columns]

    @property
    def max_column_index(self) -> int:
        return self.__columns[len(self.__columns) - 1]["index"]

    @property
    def column_count(self) -> int:
        return len(self.__columns)

    def count_of_sort_strategies(self) -> int:
        return len(self.__sort_strategies)

    def get_sort_strategies(self, enabled: bool = True) -> "List[SortStrategy]":
        result: list[SortStrategy] = []
        for sort_strategy in self.__sort_strategies:
            if sort_strategy.enabled == enabled:
                result.append(deepcopy(sort_strategy))
        result.sort(key=ExcelDefinition.__sort_priority_map, reverse=False)
        return result

    def count_of_pre_process_steps(self) -> int:
        return len(self.__pre_process_steps)

    def get_pre_process_steps(self, enabled: bool = True) -> "List[PreProcessStep]":
        result: list[PreProcessStep] = []
        for pre_process_step in self.__pre_process_steps:
            if pre_process_step.enabled == enabled:
                result.append(deepcopy(pre_process_step))
        result.sort(key=ExcelDefinition.__sort_priority_map, reverse=False)
        return result

    def get_pre_process_step_by_name(
        self, step_name: str, enabled: bool = True
    ) -> "Optional[PreProcessStep]":
        for pre_process_step in self.__pre_process_steps:
            if (
                strip_lower(pre_process_step.name) == strip_lower(step_name)
                and pre_process_step.enabled == enabled
            ):
                return deepcopy(pre_process_step)
        return None

    @staticmethod
    def __sort_priority_map(item: BasicStep) -> int:
        return item.priority

    def total_count(self):
        return len(self.__columns)

    @property
    def version(self) -> int:
        return self.__version



# -*- coding: utf-8 -*-
import pathlib
from json import loads, JSONDecodeError
from pathlib import Path
from typing import List, Tuple, Union
from .utils import strip_lower

__all__ = ["SprintScheduleStore"]


class SprintScheduleStore:
    def __init__(self) -> None:
        self.__store: List[Tuple[str, int]] = []

    def load(self, content: str):
        """
        Load json string to generate the priority list

        :param content:
            JSON string content
        """
        try:
            raw_data = loads(content)
        except JSONDecodeError as e:
            raise SyntaxError(
                f"""The structure of sprint schedule file is wrong. \
Hint: {e.msg} in line {e.lineno}:{e.colno}."""
            ) from e

        priority = -1
        sprints: List[str] = []
        for raw_item in raw_data:
            is_valid_record: bool = True
            for name, configuration in raw_item.items():
                if strip_lower(name) == strip_lower("priority"):
                    if configuration is None or not isinstance(configuration, int):
                        # Just skip invalid items.
                        is_valid_record = False
                        continue
                    priority = configuration
                if strip_lower(name) == strip_lower("sprints"):
                    if configuration is None or not isinstance(configuration, list):
                        is_valid_record = False
                        continue
                    for sprint in configuration:
                        if isinstance(sprint, str) and len(sprint) > 0:
                            sprints.append(sprint)

            if is_valid_record and priority != -1:
                for sprint in sprints:
                    self.__store.append((sprint, priority))
            sprints.clear()
            priority = -1

    def load_file(self, file: Union[str, Path]):
        """
        Load json file to generate the Excel definition

        :param file:
            JSON file location
        """

        if (
            file is None
            or not pathlib.Path(file).is_absolute()
            or not pathlib.Path(file).exists()
        ):
            raise FileNotFoundError(
                f"""Please make sure the sprint schedule file exist \
and the path should be absolute. File: {file}."""
            )

        with open(file=file, mode="r", encoding="utf-8") as schedule_file:
            try:
                self.load(schedule_file.read())
            finally:
                schedule_file.close()

    def get_priority(self, sprint: str) -> int:
        for item in self.__store:
            if sprint.upper() in item[0].upper():
                return item[1]
        return 0

    def total_count(self) -> int:
        return len(self.__store)



# -*- coding: utf-8 -*-
# pylint: disable=line-too-long
from pytest import raises

from jira_assistant.excel_definition import (
    ExcelDefinition,
    PreProcessStep,
    parse_json_item_to_pre_process_step,
    parse_json_item_to_sort_strategy,
)

from . import ASSETS_FILES


def test_pre_process_step():
    step = PreProcessStep("a", True, 1, None)
    step.name = "abc"
    assert step.name == "abc"
    step.enabled = False
    assert step.enabled is False
    step.priority = 2
    assert step.priority == 2
    step.config = {"key": "value"}
    assert step.config is not None


def test_parse_json_item_to_pre_process_step_invalid_name():
    with raises(ValueError) as err:
        parse_json_item_to_pre_process_step(
            {"Priority": 1, "Name": None, "Enabled": False, "Config": {}}
        )
    assert "The Name property in the pre-process step is invalid." in str(err.value)


def test_parse_json_item_to_pre_process_step_invalid_enabled():
    step = parse_json_item_to_pre_process_step(
        {"Priority": 1, "Name": "abc", "Enabled": "abc", "Config": {}}
    )
    assert step.enabled is False


def test_parse_json_item_to_pre_process_step_invalid_priority():
    with raises(ValueError) as err:
        parse_json_item_to_pre_process_step(
            {"Priority": None, "Name": "abc", "Enabled": False, "Config": {}}
        )
    assert "The Priority property in the pre-process step is invalid." in str(err.value)


def test_parse_json_item_to_sort_strategy_invalid_name():
    with raises(ValueError) as err:
        parse_json_item_to_sort_strategy(
            {"Priority": 1, "Name": None, "Enabled": False, "Config": {}}
        )
    assert "The Name property in the sort strategy is invalid." in str(err.value)


def test_parse_json_item_to_sort_strategy_invalid_enabled():
    strategy = parse_json_item_to_sort_strategy(
        {"Priority": 1, "Name": "abc", "Enabled": "abc", "Config": {}}
    )
    assert strategy.enabled is False


def test_parse_json_item_to_sort_strategy_invalid_priority():
    with raises(ValueError) as err:
        parse_json_item_to_sort_strategy(
            {"Priority": None, "Name": "abc", "Enabled": False, "Config": {}}
        )
    assert "The Priority property in the sort strategy is invalid." in str(err.value)


def test_load_happy_path():
    excel_definition_filename = ASSETS_FILES / "excel_definition.json"

    store = ExcelDefinition()
    with open(excel_definition_filename, encoding="utf-8") as file:
        store.load(file.read())
    assert store.total_count() > 0


def test_load_file():
    excel_definition_filename = ASSETS_FILES / "excel_definition.json"
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)
    assert store.total_count() > 0


def test_load_file_none():
    store = ExcelDefinition()
    file: str = None  # type: ignore
    with raises(FileNotFoundError) as err:
        store.load_file(file)
    assert (
        "Please make sure the excel definition file exist and the path should be absolute"
        in str(err.value)
    )


def test_iter():
    excel_definition_filename = ASSETS_FILES / "excel_definition.json"
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)
    items = []
    for item in store:
        items.append(item)
    assert len(items) > 0


def test_validate():
    excel_definition_filename = ASSETS_FILES / "excel_definition.json"
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 0


def test_validate_missing_story_id():
    excel_definition_filename = ASSETS_FILES / "excel_definition_missing_story_id.json"
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 1


def test_validate_duplicate_index():
    excel_definition_filename = ASSETS_FILES / "excel_definition_duplicate_index.json"
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 1


def test_validate_duplicate_jira_field():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_duplicate_jira_field.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 2


def test_validate_duplicate_column_name():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_duplicate_column_name.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 1


def test_validate_duplicate_inline_weights():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_duplicate_inline_weights.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 3
    assert (
        "Column do not support Inline Weights. Currently only Priority type support. Column: Entry/Last Updated Date"
        in validation_result
    )
    assert (
        "Column do not support Inline Weights. Currently only Priority type support. Column: productManager"
        in validation_result
    )
    assert (
        "Duplicate Inline Weights. Currently only support different line weights. Column: productManager"
        in validation_result
    )


def test_validate_invalid_name():
    excel_definition_filename = ASSETS_FILES / "excel_definition_invalid_name.json"
    store = ExcelDefinition()
    with raises(SyntaxError) as err:
        store.load_file(excel_definition_filename)

    assert (
        "The excel definition file has below issues need to be fixed"
        in err.value.args[0]
    )
    assert (
        "The Name property type in the column definition should be string."
        in err.value.args[0]
    )


def test_validate_invalid_raise_ranking():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_invalid_raise_ranking.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 1


def test_validate_invalid_require_sort():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_invalid_require_sort.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 2


def test_validate_invalid_type():
    excel_definition_filename = ASSETS_FILES / "excel_definition_invalid_type.json"
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 2


def test_validate_invalid_index():
    excel_definition_filename = ASSETS_FILES / "excel_definition_invalid_index.json"
    store = ExcelDefinition()
    with raises(SyntaxError) as err:
        store.load_file(excel_definition_filename)

    assert (
        "The excel definition file has below issues need to be fixed"
        in err.value.args[0]
    )
    assert (
        "The Index property type in the column definition is not integer."
        in err.value.args[0]
    )


def test_validate_index_not_continuation():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_index_not_continuation.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 1


def test_validate_strategy_duplicate_priority():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_strategy_duplicate_priority.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 1


def test_validate_strategy_invalid_name():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_strategy_invalid_name.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 1


def test_validate_invalid_json():
    excel_definition_filename = ASSETS_FILES / "excel_definition_invalid_json.txt"
    store = ExcelDefinition()

    with raises(Exception) as err:
        store.load_file(excel_definition_filename)

    assert (
        "The structure of excel definition file is wrong. Hint: Expecting ',' delimiter in line 49:1"
        in err.value.args[0]
    )


def test_validate_invalid_pre_process_step_name():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_invalid_pre_process_step_name.json"
    )
    store = ExcelDefinition()

    with raises(SyntaxError) as err:
        store.load_file(excel_definition_filename)

    assert (
        "The excel definition file has below issues need to be fixed"
        in err.value.args[0]
    )
    assert "The Name property in the pre-process step is invalid." in err.value.args[0]


def test_validate_invalid_pre_process_step_priority():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_invalid_pre_process_step_priority.json"
    )
    store = ExcelDefinition()

    store.load_file(excel_definition_filename)
    validation_result = store.validate()

    assert len(validation_result) == 1


def test_validate_strategy_duplicate_column_story_id():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_duplicate_column_story_id.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 4
    assert "Column named storyId has been duplicated." in validation_result
    assert (
        "Column do not support Inline Weights. Currently only Priority type support. Column: Entry/Last Updated Date"
        in validation_result
    )
    assert (
        "Column do not support Inline Weights. Currently only Priority type support. Column: productManager"
        in validation_result
    )
    assert (
        "Column do not support Inline Weights. Currently only Priority type support. Column: storyId"
        in validation_result
    )


def test_validate_create_jira_story_missing_project_type():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_create_jira_story_missing_project_type.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert (
        "The PreProcessStep: CreateJiraStory must have a column named ProjectType."
        in validation_result[0]
    )


def test_validate_create_jira_story_missing_issue_type():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_create_jira_story_missing_issue_type.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert (
        "The PreProcessStep: CreateJiraStory must have a column named IssueType."
        in validation_result[0]
    )


def test_validate_filter_out_story_based_on_jira_status_missing_status():
    excel_definition_filename = (
        ASSETS_FILES
        / "excel_definition_filter_out_story_based_on_jira_status_missing_status.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert (
        "The PreProcessStep: FilterOutStoryBasedOnJiraStatus must have a column named Status."
        in validation_result[0]
    )


def test_validate_filter_out_story_based_on_jira_status_missing_pre_step():
    excel_definition_filename = (
        ASSETS_FILES
        / "excel_definition_filter_out_story_based_on_jira_status_missing_pre_step.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert (
        "The step named RetrieveJiraInformation must be processed before FilterOutStoryBasedOnJiraStatus."
        in validation_result[0]
    )
